#Ref: https://ieeexplore.ieee.org/stamp/stamp.jsp?arnumber=9230365&casa_token=paW6MkN65vUAAAAA:JFJZQ3nC7fJ-6evzYq8DGWNrRAfd6qXUO
import numpy as np

from cpsim import Simulator
from cpsim.controllers.LQR import LQR

# system dynamics
kp = 2     # proportional gains of an on-board PD controller
kd = 1.5   # derivative gains of an on-board PD controller
beta = -0.1 # characterizes the loss of velocity as a result of friction
d_star = 2 # desired distance
A = np.array([[0, 0, 0, -1, 1, 0, 0],
              [0, 0, 0, 0, -1, 1, 0],
              [0, 0, 0, 0, 0, -1, 1],
              [kp, 0, 0, beta - kd, kd, 0, 0],
              [-kp, kp, 0, kd, beta - 2 * kd, kd, 0],
              [0, -kp, kp, 0, kd, beta - 2 * kd, kd],
              [0, 0, -kp, 0, 0, kd, beta - kd]])

B = np.concatenate((np.zeros((4, 3)), np.eye(4)), axis=1).T

x_0 = np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])

# utils parameters
R = np.eye(4) * 0.0008
Q = np.eye(7) * 1.2

control_limit = {
    'lo': np.array([-5]),
    'up': np.array([5])
}
class Controller:
    def __init__(self, dt, control_limit=None):
        self.lqr = LQR(A, B, Q, R)
        self.set_control_limit(control_lo=control_limit['lo'], control_up=control_limit['up'])

    def update(self, ref: np.ndarray, feedback_value: np.ndarray, current_time) -> np.ndarray:
        self.lqr.set_reference(ref)
        cin = self.lqr.update(feedback_value, current_time)
        return cin
    def set_control_limit(self, control_lo, control_up):
        self.control_lo = control_lo
        self.control_up = control_up
        self.lqr.set_control_limit(self.control_lo[0], self.control_up[0])

    def clear(self):
        self.lqr.clear()

class Platoon(Simulator):
    """
              States: (7,)
                  x[0]: e12 relative distance error with car 1 and 2
                  x[1]: e23 relative distance error with car 2 and 3
                  x[2]: e34 relative distance error with car 3 and 4
                  x[3]: velocity of car 1
                  x[4]: velocity of car 2
                  x[5]: velocity of car 3
                  x[6]: velocity of car 4
              Control Input: (4,)
                  u[0]: acceleration of car 1
                  u[1]: acceleration of car 2
                  u[2]: acceleration of car 3
                  u[3]: acceleration of car 4
                  State Feedback
              Controller: PID
              """
    def __init__(self, name, dt, max_index, noise=None):
        super().__init__('Platoon' + name, dt, max_index)
        self.linear(A, B)
        controller = Controller(dt, control_limit)
        settings = {
            'init_state': x_0,
            'feedback_type': 'state',
            'controller': controller
        }
        if noise:
            settings['noise'] = noise
        self.sim_init(settings)


if __name__ == "__main__":
    max_index = 800
    dt = 0.02
    ref = [np.array([1])] * 301 + [np.array([2])] * 300 + [np.array([1])] * 200
    noise = {
        'process': {
            'type': 'white',
            'param': {'C': np.eye(7) * 0.01}
        }
    }
    platoon = Platoon('test', dt, max_index, noise)
    for i in range(0, max_index + 1):
        assert platoon.cur_index == i
        platoon.update_current_ref(ref[i])
        # attack here
        platoon.evolve()
    # print results
    import matplotlib.pyplot as plt

    t_arr = np.linspace(0, 10, max_index + 1)
    ref = [x[0] for x in platoon.refs[:max_index + 1]]
    y_arr = [x[0] for x in platoon.outputs[:max_index + 1]]

    plt.plot(t_arr, y_arr, t_arr, ref)
    plt.show()



# Ref: K. H. Johansson, "The quadruple-tank process: a multivariable laboratory process with an adjustable zero,
# " in IEEE Transactions on Control Systems Technology, vol. 8, no. 3, pp. 456-465, May 2000, doi: 10.1109/87.845876.
# Ref: J. Giraldo, S. H. Kafash, J. Ruths and A. A. Cardenas, "DARIA: Designing Actuators to Resist Arbitrary Attacks
# Against Cyber-Physical Systems," 2020 IEEE European Symposium on Security and Privacy (EuroS&P), 2020, pp. 339-353,
# doi: 10.1109/EuroSP48549.2020.00029.

import math

import numpy as np

from cpsim import Simulator
from cpsim.controllers.PID import PID

# two operating points
op = True  # True: have minimum phase characteristics
# False: have nonminimum-phase characteristics

# parameters
A1 = A3 = 28  # Ai: cross-section of Tank i;   unit:cm^2
A2 = A4 = 32
a1 = a3 = 0.071  # cross-section of the outlet hole;  unit:cm^2
a2 = a4 = 0.057
kc = 0.5  # measured level signals are and kc*h1 and kc*h2.
g = 981  # acceleration of gravity
# parameters for corresponding operating points
h1_0 = 12.4 if op else 12.6
h2_0 = 12.7 if op else 13.0
h3_0 = 1.8 if op else 4.8
h4_0 = 1.4 if op else 4.9
v1_0 = 3.00 if op else 3.15
v2_0 = 3.00 if op else 3.15
T1 = A1 / a1 * math.sqrt(2 * h1_0 / g)  # 62 if op else 63        # Ti = Ai/ai*sqrt(2*hi_0/g)
T2 = A2 / a2 * math.sqrt(2 * h2_0 / g)  # 90 if op else 91
T3 = A3 / a3 * math.sqrt(2 * h3_0 / g)  # 23 if op else 39
T4 = A4 / a4 * math.sqrt(2 * h4_0 / g)  # 30 if op else 56
k1 = 3.33 if op else 3.14  # The voltage applied to Pump is vi and the corresponding flow is ki*vi.
k2 = 3.35 if op else 3.29
gamma1 = 0.70 if op else 0.43  # The flow to Tank 1 is gamma1*k1*v1 and the flow to Tank 4 is (1-gamma1)*k1*v1
gamma2 = 0.60 if op else 0.34

A = [[-1 / T1, 0, A3 / (A1 * T3), 0],
     [0, -1 / T2, 0, A4 / (A2 * T4)],
     [0, 0, -1 / T3, 0],
     [0, 0, 0, -1 / T4]]

B = [[gamma1 * k1 / A1, 0],
     [0, gamma2 * k2 / A2],
     [0, (1 - gamma2) * k2 / A3],
     [(1 - gamma1) * k1 / A4, 0]]

C = [[kc, 0, 0, 0],
     [0, kc, 0, 0]]

D = [[0, 0], [0, 0]]

x_0 = np.array([0, 0, 0, 0])

# utils parameters
Kp1 = 1 # 3.0
Ki1 = 1/40 #1/30
Kd1 = 0
Kp2 = 1# 2.7
Ki2 = 1/40
Kd2 = 0

control_limit = {
    'lo': np.array([0, 0]),
    'up': np.array([10, 10])
}

class Controller:
    def __init__(self, dt):
        self.dt = dt
        self.pid1 = PID(Kp1, Ki1, Kd1, current_time=-dt)
        self.pid1.setWindup(100)
        self.pid1.setSampleTime(dt)
        self.pid2 = PID(Kp2, Ki2, Kd2, current_time=-dt)
        self.pid2.setWindup(100)
        self.pid2.setSampleTime(dt)
        self.set_control_limit(control_limit['lo'], control_limit['up'])

    def update(self, ref: np.ndarray, feedback_value: np.ndarray, current_time) -> np.ndarray:
        self.pid1.set_reference(ref[0])
        cin1 = self.pid1.update(feedback_value[0], current_time)
        self.pid2.set_reference(ref[1])
        cin2 = self.pid2.update(feedback_value[1], current_time)
        return np.array([cin1, cin2])

    def set_control_limit(self, control_lo, control_up):
        self.control_lo = control_lo
        self.control_up = control_up
        self.pid1.set_control_limit(control_lo[0], control_up[0])
        self.pid2.set_control_limit(control_lo[1], control_up[1])

    def clear(self):
        self.pid1.clear(-self.dt)
        self.pid2.clear(-self.dt)


class QuadrupleTank(Simulator):
    """
    States: (4,)
        x[i-1]: hi - hi_0    i=1,2,3,4
        hi: water level of the i-th tank
    Control Input: (2,)
        u[i-1]: vi - vi_0    i=1,2
        vi: input voltages to the i-th pumps
    Output:  (2,)
        y[i]: kc*x[i]      i=0,1
        y[i]: voltages from (i+1)-th level measurement devices
        Output Feedback
    Controller: PID
    """
    def __init__(self, name, dt, max_index, noise=None):
        super().__init__('Quadruple Tank ' + name, dt, max_index)
        self.linear(A, B, C, D)
        controller = Controller(dt)
        settings = {
            'init_state': x_0,
            'feedback_type': 'output',
            'controller': controller
        }
        if noise:
            settings['noise'] = noise
        self.sim_init(settings)


if __name__ == "__main__":
    max_index = 2000
    dt = 0.1
    ref = [np.array([7, 7])]*1001 + [np.array([14, 12])]*1000
    noise = {
        'process': {
            'type': 'white',
            'param': {'C': np.eye(4) * 0.1}
        }
    }
    quadruple_tank = QuadrupleTank('test', dt, max_index, noise)
    for i in range(0, max_index + 1):
        assert quadruple_tank.cur_index == i
        quadruple_tank.update_current_ref(ref[i])
        # attack here
        quadruple_tank.evolve()
    # print results
    import matplotlib.pyplot as plt
    fig, ax = plt.subplots(2, 1)
    ax1 = ax[0]
    ax2 = ax[1]
    t_arr = np.linspace(0, 10, max_index + 1)
    ref1 = [x[0] for x in quadruple_tank.refs[:max_index + 1]]
    y1_arr = [x[0] for x in quadruple_tank.outputs[:max_index + 1]]
    ax1.set_title('x1')
    ax1.plot(t_arr, y1_arr, t_arr, ref1)
    ref2 = [x[1] for x in quadruple_tank.refs[:max_index + 1]]
    y2_arr = [x[1] for x in quadruple_tank.outputs[:max_index + 1]]
    ax2.set_title('x2')
    ax2.plot(t_arr, y2_arr, t_arr, ref2)
    plt.show()



# Ref: [1] J. M. Snider, “Automatic Steering Methods for Autonomous Automobile Path Tracking,” p. 78.

import numpy as np

from cpsim import Simulator
from cpsim.controllers.LQR import LQR

cf_ = 155494.663
cr_ = 155494.663
wheelbase_ = 2.852
steer_ratio_ = 16
steer_single_direction_max_degree_ = 470.0
mass_fl = 520
mass_fr = 520
mass_rl = 520
mass_rr = 520
mass_front = mass_fl + mass_fr
mass_rear = mass_rl + mass_rr
mass_ = mass_front + mass_rear
lf_ = wheelbase_ * (1.0 - mass_front / mass_)
lr_ = wheelbase_ * (1.0 - mass_rear / mass_)
iz_ = lf_ * lf_ * mass_front + lr_ * lr_ * mass_rear
vx = 5

A = np.zeros((4, 4), dtype=np.float32)
A[0, 1] = 1.0
A[1, 1] = -(cf_ + cr_) / mass_ / vx
A[1, 2] = (cf_ + cr_) / mass_
A[1, 3] = (lr_ * cr_ - lf_ * cf_) / mass_ / vx
A[2, 3] = 1.0
A[3, 1] = (lr_ * cr_ - lf_ * cf_) / iz_ / vx
A[3, 2] = (lf_ * cf_ - lr_ * cr_) / iz_
A[3, 3] = -1.0 * (lf_ * lf_ * cf_ + lr_ * lr_ * cr_) / iz_ / vx

B = np.zeros((4, 1), dtype=np.float32)
B[1, 0] = cf_ / mass_
B[3, 0] = lf_ * cf_ / iz_

x_0 = np.array([-0.5, 0, 0, 0])

Q = np.eye(4)
R = np.eye(1) * 10

control_limit = { 'lo': np.array([-0.261799]), 'up': np.array([0.261799])}


class Controller:
    def __init__(self, dt):
        self.lqr = LQR(A, B, Q, R)
        self.set_control_limit(control_limit['lo'], control_limit['up'])

    def update(self, ref: np.ndarray, feedback_value: np.ndarray, current_time) -> np.ndarray:
        self.lqr.set_reference(ref)
        cin = self.lqr.update(feedback_value, current_time)
        return cin

    def set_control_limit(self, control_lo, control_up):
        self.control_lo = control_lo
        self.control_up = control_up
        self.lqr.set_control_limit(self.control_lo, self.control_up)

    def clear(self):
        pass


class LaneKeeping(Simulator):
    def __init__(self, name, dt, max_index, noise=None):
        super().__init__('Lane Keeping ' + name, dt, max_index)
        self.linear(A, B)
        controller = Controller(dt)
        settings = {
            'init_state': x_0,
            'feedback_type': 'state',
            'controller': controller
        }
        if noise:
            settings['noise'] = noise
        self.sim_init(settings)


if __name__ == "__main__":
    max_index = 1000
    dt = 0.01
    ref = [np.array([0, 0, 0, 0])] * (max_index + 1)
    noise = {
        'process': {
            'type': 'white',
            'param': {'C': np.eye(4) * 0.001}
        }
    }
    lk = LaneKeeping('test', dt, max_index, noise)
    for i in range(0, max_index + 1):
        assert lk.cur_index == i
        lk.update_current_ref(ref[i])
        # attack here
        lk.evolve()
    # print results
    import matplotlib.pyplot as plt

    t_arr = np.linspace(0, 10, max_index + 1)
    ref = [x[0] for x in lk.refs[:max_index + 1]]
    y_arr = [x[0] for x in lk.outputs[:max_index + 1]]

    plt.plot(t_arr, y_arr, t_arr, ref)
    plt.show()



import numpy as np

from cpsim import Simulator
from cpsim.controllers.PID import PID

# system dynamics
R = 10000
L = 0.5
C = 0.0001

A = [[0, 1 / C], [-1 / L, -R / L]]
B = [[0], [1 / L]]
C = [[1, 0]]
D = [[0]]

x_0 = np.array([0.0, 0.0])

# utils parameters
KP = 5
KI = 5
KD = 0
control_limit = {'lo': [-15], 'up': [15]}


class Controller:
    def __init__(self, dt):
        self.dt = dt
        self.pid = PID(KP, KI, KD, current_time=-dt)
        self.pid.setWindup(100)
        self.pid.setSampleTime(dt)
        self.set_control_limit(control_limit['lo'], control_limit['up'])

    def update(self, ref: np.ndarray, feedback_value: np.ndarray, current_time) -> np.ndarray:
        self.pid.set_reference(ref[0])
        cin = self.pid.update(feedback_value[0], current_time)
        return np.array([cin])

    def set_control_limit(self, control_lo, control_up):
        self.control_lo = control_lo
        self.control_up = control_up
        self.pid.set_control_limit(self.control_lo[0], self.control_up[0])

    def clear(self):
        self.pid.clear(current_time=-self.dt)


class RlcCircuit(Simulator):
    def __init__(self, name, dt, max_index, noise=None):
        super().__init__('Aircraft Pitch ' + name, dt, max_index)
        self.linear(A, B, C)
        controller = Controller(dt)
        settings = {
            'init_state': x_0,
            'feedback_type': 'output',
            'controller': controller
        }
        if noise:
            settings['noise'] = noise
        self.sim_init(settings)


if __name__ == "__main__":
    max_index = 500
    dt = 0.02
    ref = [np.array([2])] * 201 + [np.array([3])] * 200 + [np.array([2])] * 100
    noise = {
        'process': {
            'type': 'white',
            'param': {'C': np.eye(2) * 0.01}
        }
    }
    rlc_circuit = RlcCircuit('test', dt, max_index, noise)
    for i in range(0, max_index + 1):
        assert rlc_circuit.cur_index == i
        rlc_circuit.update_current_ref(ref[i])
        # attack here
        rlc_circuit.evolve()
    # print results
    import matplotlib.pyplot as plt

    t_arr = np.linspace(0, 10, max_index + 1)
    ref = [x[0] for x in rlc_circuit.refs[:max_index + 1]]
    y_arr = [x[0] for x in rlc_circuit.outputs[:max_index + 1]]

    plt.plot(t_arr, y_arr, t_arr, ref)
    plt.show()



import matplotlib.pyplot as plt
import numpy as np

from cpsim.formal.hyperplane import Hyperplane


class Strip:
    """
    normal vector - l
    The half space is defined as a < l^T x < b
    """
    def __init__(self, l: np.ndarray, a: [int, float], b: [int, float]):
        self.l = l
        assert a < b
        self.a = a
        self.b = b
        self.dim = l.shape[0]

    def inverse_normal_vector(self):
        self.l = -self.l
        self.a, self.b = -self.b, -self.a

    # make sure l^T @ point < hp.b
    def point_to_strip(self, point):
        if self.l @ point > (self.b+self.a)/2:
            self.inverse_normal_vector()
            return True
        return False

    def center(self) -> Hyperplane:
        return Hyperplane(self.l, (self.a+self.b)/2)

    def in_strip(self, x):
        return self.a < self.l @ x < self.b

    def plot(self, x1, x2, fig=None):
        if self.dim != 2:
            return NotImplemented
        y1 = (self.b - x1 * self.l[0]) / self.l[1]
        y2 = (self.b - x2 * self.l[0]) / self.l[1]
        y3 = (self.a - x1 * self.l[0]) / self.l[1]
        y4 = (self.a - x2 * self.l[0]) / self.l[1]
        if fig is None:
            fig = plt.figure()
        X = [x1, x2]
        Y1 = [y1, y2]
        Y2 = [y3, y4]
        plt.plot(X, Y1)
        plt.plot(X, Y2)
        if fig is None:
            plt.show()
        return fig



# import mpmath as mp
import math
from statistics import NormalDist

import matplotlib.pyplot as plt
import numpy as np
from scipy.linalg import sqrtm
from scipy.stats import multivariate_normal

from cpsim.formal.half_space import HalfSpace
from cpsim.formal.strip import Strip


class GaussianDistribution:
    # only keep miu and sigma in __init__ for efficiency
    def __init__(self, miu: np.ndarray, sigma: np.ndarray):
        self.miu = miu
        self.sigma = sigma
        self.dim = miu.shape[0]

    # create from a standard normal distribution
    @classmethod
    def from_standard(cls, miu: np.ndarray, C: np.ndarray):
        sigma = C @ C.T
        return cls(miu, sigma)

    # mask method in numpy
    __array_ufunc__ = None

    # '@' overload, linear transformation
    def __rmatmul__(self, other: np.ndarray):
        miu = other @ self.miu
        sigma = other @ self.sigma @ other.T
        return GaussianDistribution(miu, sigma)

    # '+' overload,
    # 1) add a bias
    # 2) add two independent gaussian distribution with same dim
    def __add__(self, other):
        if isinstance(other, np.ndarray):
            assert self.miu.shape == other.shape
            miu = self.miu + other
            return GaussianDistribution(miu, self.sigma)
        if isinstance(other, GaussianDistribution):
            assert self.miu.shape == other.miu.shape
            miu = self.miu + other.miu
            sigma = self.sigma + other.sigma
            return GaussianDistribution(miu, sigma)
        raise NotImplemented

    def __radd__(self, other):
        return self.__add__(other)

    def __str__(self):
        string = 'Gaussian Distribution:\n  miu:' + str(self.miu)
        string += '\n  sigma:' + str(self.sigma)
        return string

    # the distribution can be transformed from standard normal distribution
    # this method returns the transformation matrix
    def transformation_matrix_from_standard(self):
        return sqrtm(self.sigma)

    # generate points from this distribution
    # each column is a point
    def random(self, size):
        return sqrtm(self.sigma) @ np.random.randn(self.dim, size) + self.miu.reshape((-1, 1))

    def prob_in_half_space(self, hs: HalfSpace):
        l = hs.l
        miu = l @ self.miu
        sigma = l @ self.sigma @ l
        # print(l, miu, sigma, hs.b)
        norm_dist = NormalDist(miu, sigma)
        P = 1-norm_dist.cdf(hs.b)
        # print(P)
        return P

    def prob_in_strip(self, s: Strip):
        l = s.l
        miu = l @ self.miu
        sigma = l @ self.sigma @ l

        # norm_dist = NormalDist(miu, sigma)
        # P = abs(norm_dist.cdf(s.b) - norm_dist.cdf(s.a))

        # update for high precision
        # mp.dps = 50
        def cdf(x, miu, sigma):
            # return 0.5 * (1.0 + mp.erf((x - miu) / (sigma * mp.sqrt(2.0))))
            return 0.5 * (1.0 + math.erf((x - miu) / (math.sqrt(sigma * 2.0))))
        P = abs(cdf(s.b, miu, sigma) - abs(cdf(s.a, miu, sigma)))

        return P

    def plot(self, x1, x2, y1, y2, fig=None):
        if self.dim != 2:
            raise NotImplemented
        x, y = np.mgrid[x1:x2:.1, y1:y2:.1]
        pos = np.empty(x.shape + (2,))
        pos[:, :, 0] = x
        pos[:, :, 1] = y
        rv = multivariate_normal(self.miu, self.sigma)
        z = rv.pdf(pos)
        if fig is None:
            fig = plt.figure()
        plt.contourf(x, y, z, levels=10, alpha=1)
        if fig is None:
            plt.show()
        return fig


if __name__ == "__main__":
    miu = np.array([1, 2])
    sigma = np.diag([2, 3])
    g1 = GaussianDistribution(miu, sigma)
    print('g1 =', g1)

    # linear transformation
    A = np.array([[1, 2], [3, 4]])
    b = np.array([-1, -1])
    g2 = A @ g1 + b
    print('A@g1+b =', g2)

    # sum of two independent ones
    g3 = g1 + g2
    print('g1+g2 =', g3)

    points = g3.random(4)
    print(points)

    norm_dist = NormalDist(2,4)



#!/usr/bin/python
#
# This file is part of IvPID.
# Copyright (C) 2015 Ivmech Mechatronics Ltd. <bilgi@ivmech.com>
#
# IvPID is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IvPID is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# title           :PID.py
# description     :python pid controller
# author          :Caner Durmusoglu
# date            :20151218
# version         :0.1
# notes           :
# python_version  :2.7
# ==============================================================================

# add utils limit - Lin
# update DTerm using delta_pv - Lin

"""Ivmech PID Controller is simple implementation of a Proportional-Integral-Derivative (PID) Controller in the Python Programming Language.
More information about PID Controller: http://en.wikipedia.org/wiki/PID_controller
"""
import time

from .controller_base import Controller


class PID(Controller):
    """PID Controller
    """

    def __init__(self, P=0.2, I=0.0, D=0.0, current_time=None):

        self.Kp = P
        self.Ki = I
        self.Kd = D

        self.sample_time = 0.00
        self.current_time = current_time if current_time is not None else time.time()
        self.last_time = self.current_time

        self.clear(current_time)
        self.control_lo = None
        self.control_up = None

    def clear(self, current_time=None):
        """Clears PID computations and coefficients"""
        self.current_time = current_time if current_time is not None else time.time()
        self.last_time = self.current_time

        self.SetPoint = 0.0

        self.PTerm = 0.0
        self.ITerm = 0.0
        self.DTerm = 0.0
        self.last_error = 0.0
        self.last_pv = None

        # Windup Guard
        self.int_error = 0.0
        # self.windup_guard = 20.0

        self.output = 0.0

    def update(self, feedback_value: float, current_time=None) -> float:
        """Calculates PID value for given reference feedback

        .. math::
            u(t) = K_p e(t) + K_i \int_{0}^{t} e(t)dt + K_d {de}/{dt}

        .. figure:: images/pid_1.png
           :align:   center

           Test PID with Kp=1.2, Ki=1, Kd=0.001 (test_pid.py)

        """
        error = self.SetPoint - feedback_value

        self.current_time = current_time if current_time is not None else time.time()
        delta_time = self.current_time - self.last_time
        delta_error = error - self.last_error

        if (delta_time >= self.sample_time):
            self.PTerm = self.Kp * error
            self.ITerm += error * delta_time

            if (self.ITerm < -self.windup_guard):
                self.ITerm = -self.windup_guard
            elif (self.ITerm > self.windup_guard):
                self.ITerm = self.windup_guard

            delta_pv = 0
            if self.last_pv:
                delta_pv = feedback_value - self.last_pv
            self.DTerm = 0.0
            if delta_time > 0:
                self.DTerm = delta_pv / delta_time

            # Remember last time and last error for next calculation
            self.last_time = self.current_time
            self.last_error = error
            self.last_pv = feedback_value

            self.output = self.PTerm + (self.Ki * self.ITerm) + (self.Kd * self.DTerm)
            if self.control_up is not None and self.control_up < self.output:
                self.output = self.control_up
            elif self.control_lo is not None and self.control_lo > self.output:
                self.output = self.control_lo
        return self.output

    def set_reference(self, ref: float):
        self.SetPoint = ref

    def setKp(self, proportional_gain):
        """Determines how aggressively the PID reacts to the current error with setting Proportional Gain"""
        self.Kp = proportional_gain

    def setKi(self, integral_gain):
        """Determines how aggressively the PID reacts to the current error with setting Integral Gain"""
        self.Ki = integral_gain

    def setKd(self, derivative_gain):
        """Determines how aggressively the PID reacts to the current error with setting Derivative Gain"""
        self.Kd = derivative_gain

    def setWindup(self, windup):
        """Integral windup, also known as integrator windup or reset windup,
        refers to the situation in a PID feedback controller where
        a large change in setpoint occurs (say a positive change)
        and the integral terms accumulates a significant error
        during the rise (windup), thus overshooting and continuing
        to increase as this accumulated error is unwound
        (offset by errors in the other direction).
        The specific problem is the excess overshooting.
        """
        self.windup_guard = windup

    def setSampleTime(self, sample_time):
        """PID that should be updated at a regular interval.
        Based on a pre-determined sampe time, the PID decides if it should compute or return immediately.
        """
        self.sample_time = sample_time

    def set_control_limit(self, control_lo: float, control_up: float):
        self.control_lo = control_lo
        self.control_up = control_up



import matplotlib.pyplot as plt
import numpy as np

from cpsim.formal.gaussian_distribution import GaussianDistribution
from cpsim.formal.strip import Strip
from cpsim.formal.zonotope import Zonotope


class ReachableSet:
    """
    get the Reachable set + Distribution
    A, B - from discrete-time system
    """

    def __init__(self, A, B, U: Zonotope, W: GaussianDistribution, max_step=50):
        self.ready = False
        self.max_step = max_step
        self.u_dim = len(U)
        self.A = A
        self.B = B
        self.A_k = [np.eye(A.shape[0])]
        for i in range(max_step):
            self.A_k.append(A @ self.A_k[-1])
        self.A_k_B = [val @ B for val in self.A_k]
        self.A_k_B_U = [val @ B @ U for val in self.A_k]
        self.A_k_W = [val @ W for val in self.A_k]
        self.bar_u_k = [U, self.A_k_B_U[0]]
        self.bar_w_k = [W, self.A_k_W[0]]
        for i in range(1, max_step):
            self.bar_u_k.append(self.A_k_B_U[i] + self.bar_u_k[-1])
            self.bar_w_k.append(self.A_k_W[i] + self.bar_w_k[-1])

    def init(self, x_0: GaussianDistribution, s: Strip):
        self.x_0 = x_0
        self.s = s
        self.hp = s.center()
        self.ready = True

    def reachable_set_wo_noise(self, k: int) -> Zonotope:
        x_0 = self.x_0.miu
        X_k = self.A_k[k] @ x_0 + self.bar_u_k[k]
        if self.s.point_to_strip(X_k.c):
            self.hp = self.s.center()
        return X_k

    def state_reconstruction(self, us) -> GaussianDistribution:
        k = len(us)
        x = self.x_0.miu
        for u in us:
            x = self.A @ x + self.B @ u
        return self.distribution(x, k)

    def first_intersection(self) -> ([int, None], Zonotope):
        for i in range(1, self.max_step):
            X_k = self.reachable_set_wo_noise(i)
            if X_k.is_intersected(self.hs):
                return i, X_k
        return None, X_k

    def distribution(self, vertex: [np.ndarray, GaussianDistribution], k: int):
        zstar_cov = self.A_k[k] @ self.x_0.sigma @ self.A_k[k].T
        zstar_distribution = GaussianDistribution(vertex, zstar_cov)
        return zstar_distribution + self.bar_w_k[k]  # count in x_0 covariance

    def reachable_set_k(self, k: int):
        X_k = self.reachable_set_wo_noise(k)
        z_star, alpha, arrive = X_k.point_closest_to_hyperplane(self.hp)
        D_k = self.distribution(z_star, k)
        P = D_k.prob_in_strip(self.s)
        return X_k, D_k, z_star, alpha, P, arrive

    def plot(self, X_k: Zonotope, D_k: GaussianDistribution, alpha, fig_setting):
        fig = plt.figure()
        if fig_setting['distribution'] and 'x1' in fig_setting and 'x2' in fig_setting and 'y1' in fig_setting and \
                'y2' in fig_setting:
            x1, x2, y1, y2 = fig_setting['x1'], fig_setting['x2'], fig_setting['y1'], fig_setting['y2']
            D_k.plot(x1, x2, y1, y2, fig)
        if fig_setting['zonotope']:
            X_k.plot(fig)
        if fig_setting['strip'] and 'x1' in fig_setting and 'x2' in fig_setting:
            self.s.plot(fig_setting['x1'], fig_setting['x2'], fig)
        if fig_setting['routine']:
            head_width = line_width = None
            if 'head_width' in fig_setting:
                head_width = fig_setting['head_width']
                line_width = fig_setting['width']
            X_k.show_control_effect(alpha, self.u_dim, head_width, line_width, fig)
        if 'x1' in fig_setting and 'x2' in fig_setting:
            plt.xlim((fig_setting['x1'], fig_setting['x2']))
        if 'y1' in fig_setting and 'y2' in fig_setting:
            plt.ylim((fig_setting['y1'], fig_setting['y2']))
        plt.show()

    def given_k(self, max_k: int):
        if not self.ready:
            print('Init before recovery!')
            raise RuntimeError
        dummy_res = (None, None, None, None, 0, False)
        reach_res = [dummy_res]
        arrived = False
        max_P = 0
        for i in range(1, max_k + 1):
            res = self.reachable_set_k(i)
            reach_res.append(res)
            # X_k, D_k, z_star, alpha, P, arrive = res
            if res[4] > max_P:
                max_P = res[4]
            if res[5] and not arrived:
                arrived = True
                break
        all_res = [val for val in reach_res if val[4] == max_P]
        res = all_res[-1]
        k = reach_res.index(res)
        return k, *res

    def given_P(self, P_given: float, max_k: int):
        if not self.ready:
            print('Init before recovery!')
            raise RuntimeError
        satisfy = False
        i = 0
        X_k = D_k = z_star = alpha = P = arrive = None
        for i in range(1, max_k + 1):
            X_k, D_k, z_star, alpha, P, arrive = self.reachable_set_k(i)
            if P > P_given:
                satisfy = True
                break
            if arrive == True:
                break
        return i, satisfy, X_k, D_k, z_star, alpha, P, arrive

    def maintain_once(self, P_given: float):
        if not self.ready:
            print('Init before recovery!')
            raise RuntimeError
        res = self.reachable_set_k(1)
        if res[4] >= P_given:
            return True, *res
        else:
            return False, *res


if __name__ == '__main__':
    u_lo = np.array([-1, -2])
    u_up = np.array([3, 4])
    U = Zonotope.from_box(u_lo, u_up)
    print(U)
    # U.plot()

    A = np.array([[1, 1], [0, 2]])
    B = np.array([[2, 0], [0, 1]])
    W = GaussianDistribution.from_standard(miu=np.array([0, 0]), C=np.diag([0.3, 0.3]))
    reach = ReachableSet(A, B, U, W, max_step=5)
    x_0 = GaussianDistribution(np.array([5, 5]), np.eye(2))

    s = Strip(np.array([1, 1]), a=100, b=120)
    reach.init(x_0, s)

    fig_setting = {'x1': 0, 'x2': 80, 'y1': 0, 'y2': 90,
                   'strip': True, 'routine': True,
                   'zonotope': True, 'distribution': True}
    X_k, D_k, z_star, alpha, P, arrive = reach.reachable_set_k(1)
    # reach.plot(X_k, D_k, alpha, fig_setting)
    print('i=', 1, 'P=', P, 'z_star=', z_star, 'arrive=', arrive, 'alpha=', alpha)

    X_k, D_k, z_star, alpha, P, arrive = reach.reachable_set_k(2)
    # reach.plot(X_k, D_k, alpha, fig_setting)
    print('i=', 2, 'P=', P, 'z_star=', z_star, 'arrive=', arrive)

    X_k, D_k, z_star, alpha, P, arrive = reach.reachable_set_k(3)
    # reach.plot(X_k, D_k, alpha,fig_setting)
    print('i=', 3, 'P=', P, 'z_star=', z_star, 'arrive=', arrive)

    i, satisfy, X_k, D_k, z_star, alpha, P, arrive = reach.given_P(P_given=0.9, max_k=10)
    print('i=', i, 'found=', satisfy)
    print('i=', i, 'P=', P, 'z_star=', z_star, 'arrive=', arrive)
    print('alpha=', alpha)
    # reach.plot(X_k, D_k, alpha, fig_setting)
    rec_u = U.alpha_to_control(alpha)
    print('rec_u =', rec_u)

    x1 = A @ np.array([5, 5]) + B @ rec_u[0]
    x2 = A @ x1 + B @ rec_u[1]
    x3 = A @ x2 + B @ rec_u[1]
    print('x3 =', x3)

    # k, X_k, D_k, z_star, alpha, P, arrive = reach.given_k(5)
    # print('i=', i, 'P=', P, 'z_star=', z_star, 'arrive=', arrive)
    # reach.plot(X_k, D_k, alpha, fig_setting)



import numpy as np


class Hyperplane:
    def __init__(self, l: np.ndarray, b: [int, float]):
        self.l = l
        self.b = b
        self.dim = l.shape[0]

    def inverse_normal_vector(self):
        self.l = -self.l
        self.b = -self.b



from interval import interval

DISTURBANCE_INTERVAL = interval[0, 0] # interval[-0.01, 0.01]

# Start helper functions
def convert_1D_list_to_intervals(list_1D):
    new_list_1D = []
    for element in list_1D:
        new_list_1D.append(interval([element, element]))
    return new_list_1D

def convert_2D_list_to_intervals(list_2D):
    new_list_2D = []
    for vector in list_2D:
        new_vector = []
        for element in vector:
            new_vector.append(interval([element, element]))
        new_list_2D.append(new_vector)
    return new_list_2D

def replace_safe_state_intervals_with_saved_value_intervals(x_interval, x_saved_interval, unsafeSensors_oneHot):
    for i, unsafe in enumerate(unsafeSensors_oneHot):
        if not unsafe:
            x_interval[i] = x_saved_interval[i]
    return x_interval

# Helper functions for operations on lists of intervals
def __mul__(list_of_intervals, constant):
    return [Interval*constant for Interval in list_of_intervals]

def __add__(list_of_intervals, another_list_of_intervals):
    assert len(list_of_intervals)==len(another_list_of_intervals), "You can't add 2 lists (of intervals) with different number of elements"
    new_list = []
    for i in range(len(list_of_intervals)):
        new_list.append(list_of_intervals[i]+another_list_of_intervals[i])
    return new_list

def __sum__(list_of_list_of_intervals):
    summed_list = list_of_list_of_intervals[0]
    for list_of_intervals in list_of_list_of_intervals[1:]:
        for i, interval in enumerate(list_of_intervals):
            summed_list[i] += interval
    return summed_list

def __add_to_every_element__(list_of_intervals, another_interval):
    return [Interval+another_interval for Interval in list_of_intervals]

# Helper functions for numerical integration
def RungeKutta_Integration(dynamics, x_interval, u_interval, step_size):
    x = x_interval
    u = u_interval
    t = None

    k_1 = dynamics(t, x, u)

    k_2 = dynamics(t, __add__(x, __mul__(k_1, step_size/2.0)) , u)

    k_3 = dynamics(t, __add__(x, __mul__(k_2, step_size/2.0)) , u)

    k_4 = dynamics(t, __add__(x, __mul__(k_3, step_size)) , u)

    x_next = __add__(x, __sum__( [__mul__(k_1, step_size/6.0), __mul__(k_2, 2.0*step_size/6.0), __mul__(k_3, 2.0*step_size/6.0), __mul__(k_4, step_size/6.0)] ))
    
    return __add_to_every_element__(x_next, DISTURBANCE_INTERVAL)

def Newton_Integration(dynamics, x_interval, u_interval, step_size):
    x = x_interval
    u = u_interval
    t = None
    x_dot = dynamics(t, x, u)
    x_next = __add__(x, __mul__(x_dot, step_size))
    
    return __add_to_every_element__(x_next, DISTURBANCE_INTERVAL)

# Helper functions for midpoint and intersection
def isIntersecting(Int1, Int2):
    if len(Int1 & Int2)==0:
        return False
    else:
        return True

def midpoint(Int):
    lower_bound = Int[0][0]
    upper_bound = Int[0][1]
    return (lower_bound+upper_bound)/2.0

# reachability
def reach(dynamics, x0_interval, Us_saved, Xs_saved, unsafeSensors_oneHot, step_size):
    if Xs_saved:
        assert len(Us_saved) == len(Xs_saved)-1, "There should be one less u vector than saved x vectors! ("
    
    total_steps = len(Us_saved)
    x_interval = x0_interval
    if Xs_saved:
        Xs_saved_interval = convert_2D_list_to_intervals(Xs_saved)
    Us_saved_interval = convert_2D_list_to_intervals(Us_saved)
    for i in range(total_steps):
        if  Xs_saved:
            x_interval = replace_safe_state_intervals_with_saved_value_intervals(x_interval, Xs_saved_interval[i], unsafeSensors_oneHot)
        u_interval = Us_saved_interval[i]
        
        # x_interval = Newton_Integration(dynamics, x_interval, u_interval, step_size)
        x_interval = RungeKutta_Integration(dynamics, x_interval, u_interval, step_size)
        #print(x_interval)
    return x_interval

if __name__ == '__main__':
    import sys
    sys.path.append('../../../')
    from cpsim.models.nonlinear.continuous_stirred_tank_reactor import cstr_imath as cstr
    
    # This code won't run if this file is imported.
    # print(interval([1,1]) == interval([1])) # True
    x0_interval = [interval[1,1], interval[3,3]]
    #print(x0_interval)
    Us_saved = [ [0], [0.1], [1]]
    Xs_saved = [ [4,5], [6,7], [8,9], [10,11]]
    unsafeSensors_oneHot = [0,1]
    step_size = 0.1
    out = reach(cstr, x0_interval, Us_saved, Xs_saved, unsafeSensors_oneHot, step_size)
    
    print(f'{out=}')
    print(f'{midpoint(interval[5,9])=}')



from math import cos, sin, copysign, pi

import numpy as np


from cpsim import Simulator
from cpsim.controllers.PID import PID


#Helper Function
def kn2ms(x):
    return 1852*x/3600.0

def ms2kn(x):
    return 3600*x/1852.0

def nmi2m(x):
    return 1852*x

def m2nmi(x):
    return x/1852.0

def deg2rad(x):
    return x*pi/180

def rad2deg(x):
    return x*180/pi

def saturate(x, a, b, use_imath=False):
    if use_imath:
        from interval import imath, interval
        if b < x[0][0]:
            min_of_b_and_x = interval([b, b])
        else:
            min_of_b_and_x = x

        if a > min_of_b_and_x[0][1]:
            max_of_a_and_above = interval([a, a])
        else:
            max_of_a_and_above = x 

        return max_of_a_and_above

        # return imath.max(a, imath.min(x, b))
    else:
        return max(a, min(x, b))

def heading_circle(x):
    if x > pi:
        return x-2*pi
    elif x < -pi:
        return x+2*pi
    return x
#Initial state
x_0 = np.array([0, 0, 0, 0, 0, 0, 0, 0])

# utils parameters
Kp1 = 4.5e-1 # 3.0
Ki1 = 5e-2#1/30
Kd1 = 0
Kp2 = 1e-1# 2.7
Ki2 = 0
Kd2 = -3.5e0



control_limit = {
    'lo': np.array([0, -1]),
    'up': np.array([1, 1])
}
# print(heading_circle(deg2rad(45)))
#ODE
def vessel(x, u, use_imath=False):
    if use_imath:
        cs, ss = imath.cos(x[2]), imath.sin(x[2])
    else:
        cs, ss = cos(x[2]), sin(x[2])
    #parameters
    m = 127.92     # Mass, kg
    Xg = 0     # X coordinate for center of gravity, m
    Iz = 61.967     # Moment of inertia for Z axis (i.e., heave axis), kg.m^2
	    
    # Linear damping parameters
    # Xu = 0
    # Yv = 0
    Yr = 0
    Nv = 0
    # Nr = 0
    Xu = -2.332
    Yv = -4.673
    Nr = -0.01675


	# Inertia matrix parameters

    # Xu_dot = 0
    # Yv_dot = 0
    # Yr_dot = 0
    # Nv_dot = 0
    # Nr_dot = 0
    Xu_dot = 3.262
    Yv_dot = 28.89
    Yr_dot = 0.525
    Nv_dot = 0.157
    Nr_dot = 13.98
	# Restorative forces parameters
    # Xx = 0
    # Yy = 0
    # Npsi = 0
    # Helper variables
    a1 = Iz*Yv_dot - Nr_dot*Yv_dot + Nv_dot*Yr_dot - Iz*m + Nr_dot*m + (Xg*m)**2 - Nv_dot*Xg*m - Xg*Yr_dot*m
    a2 = u[1] + Yr*x[2] * Yv*x[1]
    a3 = u[2] + Nr*x[2] * Nv*x[1]

    # State transition functions
    f1 = x[3]*cs - x[4]*ss
    f2 = x[3]*ss + x[4]*cs
    f3 = x[5]
    f4 = (u[0] + Xu*x[3])/(m-Xu_dot)
    f5 = (1/a1)*( (Xg*m - Yr_dot)*a3 + (Nr_dot - Iz)*a2 )
    f6 = (1/a1)*( (m - Yv_dot)*a3 + (m*Xg - Nv_dot)*a2 )

    # States derivative
    return [f1, f2, f3, f4, f5, f6]

def rudder(command, surge_v, use_imath=False):
    rudder_v = 0.166
    rudder_r = 1.661
    rudder_max_angle = deg2rad(30)
    angle = saturate(command, -rudder_max_angle, rudder_max_angle, use_imath)
    return -rudder_v * surge_v * angle, -rudder_r * surge_v * angle

def copysign_custom(a, b, use_imath=False):
    if use_imath:
        return copysign(a, b[0][0])
    else:
        return copysign(a, b)

def propulsion(torque, shaft_speed, use_imath=False):
    Is = 25000.0/747225.0    # Moment of inertia
    Kq = 0.0113   # Load torque coefficient
    Kt = 0.3763   # Load thrust coefficient
    D = 0.3
    max_shaft_speed = 157
    rho = 1000
    this_shaft = (1/Is)*( torque - copysign_custom(1, shaft_speed, use_imath)*Kq*rho*D**5*(shaft_speed/(2*pi))**2 )
    omega = saturate(shaft_speed, -max_shaft_speed, max_shaft_speed, use_imath)
    return copysign_custom(1, omega, use_imath) * Kt * rho * D**4 * (omega/(2*pi))**2, this_shaft

def dcmotor(power, current, omega, use_imath=False):
    K = 0.3                         # Torque and back EMF constant
    R = 1                         # Armature resistance
    L = 0.5                         # Armature inductance
    efficiency = 0.8      # Motor efficiency (i.e., yield)
    max_current = 22
    this_current = (power- R*current - K*omega)/L
    I = saturate(current, -max_current, max_current, use_imath)
    return (K*I)*efficiency, this_current

def battery(signal):
	voltage = 12
	return voltage*signal


def whole_model(t, x, u, use_imath=False):
	# state 0-5 for vessel (East, North, Yaw, and their velocities)
	# state 6 for propeller Angular shaft speed
	# state 7 for motor current

	# two inputs: The first is for the battery, the second is for the rudder command

    power = battery(u[0])

    torque, current = dcmotor(power, x[7], x[6], use_imath)
    thrust, shaft = propulsion(torque, x[6],use_imath)
    f_sway, f_yaw = rudder(u[1], x[3], use_imath)
    first_6 = vessel(x[:6], [thrust, f_sway, f_yaw], use_imath)
    if use_imath:
        first_6.append(shaft)
        first_6.append(current)
    else:
        first_6.append(shaft)
        first_6.append(current)
        first_6 = np.array(first_6)
    return first_6
def whole_model_imath(t,x,u):
    return whole_model(t=t,x=x,u=u,use_imath=True)




class Controller:
    def __init__(self, dt):
        self.dt = dt
        self.pid1 = PID(Kp1, Ki1, Kd1, current_time=-dt)
        self.pid1.setWindup(100)
        self.pid1.setSampleTime(dt)
        self.pid2 = PID(Kp2, Ki2, Kd2, current_time=-dt)
        self.pid2.setWindup(100)
        self.pid2.setSampleTime(dt)
        self.set_control_limit(control_limit['lo'], control_limit['up'])
        self.pid2.setWindup(deg2rad(30))

    def update(self, ref: np.ndarray, feedback_value: np.ndarray, current_time) -> np.ndarray:
        self.pid1.set_reference(ref[3])
        cin1 = self.pid1.update(feedback_value[3], current_time)
        self.pid2.set_reference(ref[2])
        cin2 = self.pid2.update(feedback_value[2], current_time)
        return np.array([cin1, cin2])

    def set_control_limit(self, control_lo, control_up):
        self.control_lo = control_lo
        self.control_up = control_up
        self.pid1.set_control_limit(control_lo[0], control_up[0])
        self.pid2.set_control_limit(control_lo[1], control_up[1])

    def clear(self):
        self.pid1.clear(-self.dt)
        self.pid2.clear(-self.dt)





class VESSEL(Simulator):
    """
        States: (2,)
            x[0]: Concentration of A in CSTR (mol/m^3)
            x[1]: Temperature in CSTR (K)
        Control Input: (1,)
            u[0]: Temperature of cooling jacket (K)
        Output:  (2,)
            State Feedback
        Controller: PID
    """

    def __init__(self, name, dt, max_index, noise=None):
        super().__init__('Vessel ' + name, dt, max_index)
        self.nonlinear(ode=whole_model, n=8, m=2, p=8)  # number of states, utils inputs, outputs
        controller = Controller(dt)
        settings = {
            'init_state': x_0,
            'feedback_type': 'state',  # 'state' or 'output',  you must define C if 'output'
            'controller': controller
        }
        if noise:
            settings['noise'] = noise
        self.sim_init(settings)


if __name__ == "__main__":
    max_index = 500
    dt = 1
    ref = [np.array([0, 0, heading_circle(deg2rad(90)), kn2ms(1), 0,0,0,0])] * (max_index+1)
    # ref = [np.array([0, 0, 45, 1, 0,0,0,0])] * (max_index+1)
    # bias attack example
    from cpsim import Attack
    bias = np.array([0, 0, 0, 0, 0, 0, 0, 0])
    bias_attack = Attack('bias', bias, 300)
    ip = VESSEL('test', dt, max_index)
    for i in range(0, max_index + 1):
        assert ip.cur_index == i
        ip.update_current_ref(ref[i])
        # attack here
        ip.cur_feedback = bias_attack.launch(ip.cur_feedback, ip.cur_index, ip.states)
        ip.evolve()

    # print results
    import matplotlib.pyplot as plt

    fig, ax = plt.subplots(6, 1)
    ax1, ax2, ax3, ax4, ax5, ax6 = ax
    t_arr = np.linspace(0, 5, max_index + 1)
    ref1 = [x[3] for x in ip.refs[:max_index + 1]]
    y1_arr = [x[3] for x in ip.outputs[:max_index + 1]]
    ax1.set_title('Speed')
    ax1.plot(t_arr, y1_arr, t_arr, ref1)
    ref2 = [x[2] for x in ip.refs[:max_index + 1]]
    y2_arr = [x[2] for x in ip.outputs[:max_index + 1]]
    ax2.set_title('Yaw')
    ax2.plot(t_arr, y2_arr, t_arr, ref2)
    y3_arr = [x[1] for x in ip.outputs[:max_index + 1]]
    ax3.set_title('East')
    ax3.plot(t_arr, y3_arr)
    y4_arr = [x[0] for x in ip.outputs[:max_index + 1]]
    ax4.set_title('Position')
    ax4.plot(y3_arr, y4_arr)
    y5_arr = [x[1] for x in ip.inputs[:max_index + 1]]
    ax5.set_title('Inputs')
    ax5.plot(t_arr, y5_arr)
    y6_arr = [x[0] for x in ip.inputs[:max_index + 1]]
    ax6.set_title('Inputs')
    ax6.plot(t_arr, y6_arr)
    plt.show()


# Ref:

import numpy as np

from cpsim import Simulator
from cpsim.controllers.PID import PID

# system dynamics
state_num = 45
A = np.zeros((state_num, state_num))
tmp = np.array([[1], [-2], [1]])
A[0:2, 0:1] = tmp[1:3, 0:1]
for i in range(state_num - 2):
    A[i:i + 3, i + 1:i + 2] = tmp
A[state_num - 2:state_num, state_num - 1:state_num] = tmp[0:2, 0:1]

if state_num == 2:
    A = np.array([[-2, 1], [1, -2]])

B = np.zeros((state_num, 1))
u_point = (state_num + 1) // 3 - 1
B[u_point, 0] = 1

C = np.zeros((1, state_num))
y_point = (state_num + 1) // 3 * 2 - 1
C[0, y_point] = 1

D = np.zeros((1, 1))

x_0 = np.zeros((state_num))

# utils parameters
# KP = 0.00000099 * state_num * state_num * state_num * state_num
# KI = 0
# KD = 0
KP = 10
KI = 0.23
KD = 0
control_limit = {'lo': [-0.5], 'up': [50]}
R = np.eye(1) * 0.0008
Q = np.eye(state_num)


class Controller:
    def __init__(self, dt):
        self.dt = dt
        self.pid = PID(KP, KI, KD, current_time=-dt)
        self.pid.setWindup(100)
        self.pid.setSampleTime(dt)
        self.set_control_limit(control_limit['lo'], control_limit['up'])

    def update(self, ref: np.ndarray, feedback_value: np.ndarray, current_time) -> np.ndarray:
        self.pid.set_reference(ref[0])
        cin = self.pid.update(feedback_value[0], current_time)
        return np.array([cin])

    def set_control_limit(self, control_lo, control_up):
        self.control_lo = control_lo
        self.control_up = control_up
        self.pid.set_control_limit(self.control_lo[0], self.control_up[0])

    def clear(self):
        self.pid.clear(current_time=-self.dt)
# class Controller:
#     def __init__(self, dt, control_limit=None):
#         self.lqr = LQR(A, B, Q, R)
#         self.lqr.set_control_limit(control_limit['lo'], control_limit['up'])
#
#     def update(self, ref: np.ndarray, feedback_value: np.ndarray, current_time) -> np.ndarray:
#         self.lqr.set_reference(ref)
#         cin = self.lqr.update(feedback_value, current_time)
#         return cin


class Heat(Simulator):
    """
                States: (N,)
                   x[i]: temperature on the i'th field
                Control Input: (1,)
                    heat source to heat the a point of the rod located at 1/3 of the length
                Output:  (1,)
                    record the temperature at 2/3 of the length
                    Output Feedback
                Controller: PID
                """
    def __init__(self, name, dt, max_index, noise=None):
        super().__init__('Aircraft Pitch ' + name, dt, max_index)
        self.linear(A, B)
        controller = Controller(dt)
        settings = {
            'init_state': x_0,
            'feedback_type': 'state',
            'controller': controller
        }
        if noise:
            settings['noise'] = noise
        self.sim_init(settings)


if __name__ == "__main__":
    max_index = 1000
    dt = 1
    ref = [np.array([15])] * (max_index + 1)
    noise = {
        'process': {
            'type': 'white',
            'param': {'C': np.eye(state_num) * 0.01}
        }
    }
    heat = Heat('test', dt, max_index, noise)
    for i in range(0, max_index + 1):
        assert heat.cur_index == i
        heat.update_current_ref(ref[i])
        # attack here
        heat.evolve()
    # print results
    import matplotlib.pyplot as plt

    t_arr = np.linspace(0, 10, max_index + 1)
    ref = [x[0] for x in heat.refs[:max_index + 1]]
    y_arr = [x[0] for x in heat.outputs[:max_index + 1]]

    plt.plot(t_arr, y_arr, t_arr, ref)
    plt.show()

    # u_arr = [x[0] for x in heat.inputs[:max_index + 1]]
    # plt.plot(t_arr, u_arr)
    # plt.show()


import logging
import os
import sys
import time
from copy import deepcopy

import numpy as np
from scipy.integrate import solve_ivp
from scipy.signal import StateSpace

from cpsim.formal.gaussian_distribution import GaussianDistribution
from cpsim.utils.linearizer import Linearizer

logging.basicConfig(
    format="%(asctime)s [%(levelname)s] %(message)s",
    handlers=[
        logging.FileHandler("debug.log"),
        logging.StreamHandler(sys.stdout)
    ]
)
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

rseed_str = os.getenv('RANDOM_SEED')
if rseed_str is None:
    rseed = np.uint32(int(time.time()))
else:
    try:
        rseed = np.uint32(rseed_str)
    except Exception as e:
        print('rseed read error:', e)
        rseed = np.uint32(int(time.time()))
logger.info(f'simulator: numpy random seed is set to {rseed}.')
np.random.seed(rseed)

class Simulator:
    """
    states, utils inputs/outputs are instance of np.array with shape (n,) (m,) (p,)
    """

    def __init__(self, name, Ts, max_index):
        self.name = name
        self.model_type = None
        self.dt = Ts
        self.sysc = None
        self.sysd = None
        self.max_index = max_index
        self.C = None
        self.D = None
        self.n = None  # number of states
        self.m = None  # number of utils inputs
        self.p = None  # number of sensor measurements
        self.ode = None  # ode function
        self.init_state = None
        # values under self.cur_index
        self.cur_x = None
        self.cur_y = None
        self.cur_u = None
        self.cur_feedback = None
        self.cur_ref = None
        self.cur_index = 0
        self.noise_setting = None
        self.m_noise = None
        self.p_noise = None
        self.m_noise_dist = None
        self.p_noise_dist = None

    def data_init(self):
        self.inputs = np.empty((self.max_index + 2, self.m), dtype=np.float)
        self.outputs = np.empty((self.max_index + 2, self.p), dtype=np.float)
        self.states = np.empty((self.max_index + 2, self.n), dtype=np.float)
        if self.feedback_type == 'output':
            self.feedbacks = np.empty((self.max_index + 2, self.p), dtype=np.float)
            self.refs = np.empty((self.max_index + 2, self.p), dtype=np.float)
        elif self.feedback_type == 'state':
            self.feedbacks = np.empty((self.max_index + 2, self.n), dtype=np.float)
            self.refs = np.empty((self.max_index + 2, self.n), dtype=np.float)  # reference value

    def reset(self):
        # the noise will not reset!
        self.data_init()
        self.cur_index = 0
        self.set_init_state(self.init_state)
        self.controller.clear()

    def linear(self, A, B, C=None, D=None):
        self.model_type = 'linear'
        if C is None:
            C = np.eye(len(A))
        if D is None:
            D = np.zeros((len(C), len(B[0])))
        self.sysc = StateSpace(A, B, C, D)
        self.sysd = self.sysc.to_discrete(self.dt)
        self.n = self.sysc.A.shape[1]
        self.m = self.sysc.B.shape[1]
        self.p = self.sysc.C.shape[0]
        self.C = self.sysc.C
        self.D = self.sysc.D

        def fprime(t, x, u):
            return self.sysc.A @ x + self.sysc.B @ u

        self.ode = fprime

    def nonlinear(self, ode, n, m, p, C=None, D=None):
        self.model_type = 'nonlinear'
        self.C = np.eye(n) if C is None else np.array(C)
        self.D = np.zeros((p, m)) if D is None else np.array(D)
        self.n = n
        self.m = m
        self.p = p
        self.ode = ode

    def linearize_at(self, x_0: np.ndarray, u_0: np.ndarray):
        """
        self.sysc = Ax + Bu + c
        """
        assert self.model_type == 'nonlinear'
        linearize = Linearizer(self.ode, self.n, self.m)
        Ac, Bc, Cc = linearize.at(x_0, u_0)
        self.sysc = StateSpace(Ac, Bc, self.C, self.D)
        self.sysc.c = Cc
        self.sysd = self.sysc.to_discrete(self.dt)
        self.sysd.c = self.dt * Cc

    def sim_init(self, settings: dict):
        """
        keys:
          'feedback_type': 'state', 'output', None
          'init_state': np.ndarray   (n,)
          'controller':  object with update method
        """
        self.set_feedback_type(settings['feedback_type'])
        self.data_init()
        if 'noise' in settings:
            self.noise_init(settings['noise'])
        self.set_init_state(settings['init_state'])
        self.set_controller(settings['controller'])

    def noise_init(self, noise):
        """
        Only implement the white noise
        keys:
          'process'/'measurement':
            'type': 'white'  todo: 'white_bounded', 'box_uniform', 'ball_uniform'
            'param':
              'C': linear transformation matrix from standard normal distribution      scale for 'white'
        """
        if 'process' in noise:
            if noise['process']['type'] == 'white':
                miu = np.zeros((self.n,))
                C = noise['process']['param']['C']
                self.p_noise_dist = GaussianDistribution.from_standard(miu, C)
                self.p_noise = self.p_noise_dist.random(self.max_index + 2).T
            elif noise['process']['type'] == 'box_uniform':
                lo = noise['process']['param']['lo']
                up = noise['process']['param']['up']
                assert up.shape == (self.n,) and lo.shape == (self.n,)
                noise_base = np.random.random_sample((self.max_index + 2, self.n))
                self.p_noise = lo + noise_base * (up-lo)

        if 'measurement' in noise:
            if noise['measurement']['type'] == 'white':
                miu = np.zeros((self.p,))
                C = noise['measurement']['param']['C']
                self.m_noise_dist = GaussianDistribution.from_standard(miu, C)
                self.m_noise = self.m_noise_dist.random(self.max_index + 2).T
            elif noise['measurement']['type'] == 'box_uniform':
                lo = noise['process']['param']['lo']
                up = noise['process']['param']['up']
                assert up.shape == (self.p,) and lo.shape == (self.p,)
                noise_base = np.random.random_sample((self.max_index + 2, self.p))
                self.p_noise = lo + noise_base * (up - lo)

    def set_init_state(self, x):
        self.init_state = x
        self.cur_x = x
        self.cur_y = self.C @ self.cur_x
        if self.m_noise is not None:
            self.cur_y += self.m_noise[self.cur_index]
        self.outputs[0] = self.cur_y
        self.states[0] = self.cur_x
        if self.feedback_type:
            self.cur_feedback = self.cur_x if self.feedback_type == 'state' else self.cur_y

    def set_feedback_type(self, feedback_type):
        """
        'state', 'output', None
        """
        self.feedback_type = feedback_type

    def set_controller(self, controller):
        """
        please implement update method to get utils input
        """
        self.controller = controller

    def update_current_ref(self, ref):
        self.cur_ref = ref

    def evolve(self, u=None):
        # record data
        self.feedbacks[self.cur_index] = deepcopy(self.cur_feedback)
        self.refs[self.cur_index] = deepcopy(self.cur_ref)

        # compute utils input
        if self.feedback_type:
            self.cur_u = self.controller.update(self.cur_ref, self.cur_feedback, self.dt * self.cur_index)
        else:
            self.cur_u = u
        # override utils input
        if not (u is None):
            self.cur_u = u
        assert self.cur_u.shape == (self.m,)
        self.inputs[self.cur_index] = deepcopy(self.cur_u)

        # implement utils input
        ts = (self.cur_index * self.dt, (self.cur_index + 1) * self.dt)
        res = solve_ivp(self.ode, ts, self.cur_x, args=(self.cur_u,))
        self.cur_index += 1
        self.cur_x = res.y[:, -1]
        if self.model_type == 'linear':
            self.cur_x = self.sysd.A @ self.cur_x + self.sysd.B @ self.cur_u
        if self.p_noise is not None:  # process noise
            self.cur_x += self.p_noise[self.cur_index]
        self.cur_y = self.C @ self.cur_x + self.D @ self.cur_u
        if self.m_noise is not None:  # measurement noise
            self.cur_y += self.m_noise[self.cur_index]
        assert self.cur_x.shape == (self.n,)
        assert self.cur_y.shape == (self.p,)
        self.states[self.cur_index] = deepcopy(self.cur_x)
        self.outputs[self.cur_index] = deepcopy(self.cur_y)

        # prepare feedback
        if self.feedback_type:
            self.cur_feedback = self.cur_x if self.feedback_type == 'state' else self.cur_y
            # self.cur_feedback may be attacked before implement
        else:
            self.cur_feedback = None

        return self.cur_index



from typing import Union

import cvxpy as cp
import numpy as np

from cpsim.controllers.controller_base import Controller


class LP(Controller):
    def __init__(self, param_dict: dict):
        assert 'Bd' in param_dict
        self.update_params(**param_dict)

    def update_params(self, **kwargs):
        """
        parameter template:
        lp_settings = {
            'Ad': ,  'Bd': ,
            'N': ,
            'ddl': , 'target_lo': , 'target_up': ,
            'safe_lo': , 'safe_up': ,
            'control_lo': , 'control_up': ,
            'ref':
        }
        """
        if 'Ad' in kwargs and 'Bd' in kwargs:
            self.update_model(kwargs['Ad'], kwargs['Bd'])
        if 'N' in kwargs:
            self.update_horizon(kwargs['N'])
        if 'ddl' in kwargs:
            self.update_ddl(kwargs['ddl'])
        if 'target_lo' in kwargs and 'target_up' in kwargs:
            self.update_target_set(kwargs['target_lo'], kwargs['target_up'])
        if 'safe_lo' in kwargs and 'safe_up' in kwargs:
            self.update_safe_set(kwargs['safe_lo'], kwargs['safe_up'])
        if 'control_lo' in kwargs and 'control_up' in kwargs:
            self.set_control_limit(kwargs['control_lo'], kwargs['control_up'])
        if 'ref' in kwargs:
            self.set_reference(kwargs['ref'])
        if 'solver' in kwargs:
            self.set_solver(kwargs['solver'])

        if 'c_nonlinear' in kwargs:
            self.update_model_residual(kwargs['c_nonlinear'])
        else:
            self.update_model_residual(None)

    def ready_to_formulate(self):
        required = ['Ad', 'Bd', 'N', 'xr', 'x0']
        ddl_required = ['xtmin', 'xtmax']
        short = []
        for key in required:
            if not hasattr(self, key):
                short.append(key)
        if hasattr(self, 'ddl'):
            for key in ddl_required:
                if not hasattr(self, key):
                    short.append(key)
        # default values
        if not hasattr(self, 'xmin'):
            self.update_safe_set(np.array([-np.inf] * self.nx), np.array([np.inf] * self.nx))
        if not hasattr(self, 'umin'):
            self.set_control_limit(np.array([-np.inf] * self.nu), np.array([np.inf] * self.nu))
        if len(short) == 0:
            return True
        else:
            print(short, 'are required but not provided.')
            return False

    def update_model(self, Ad: np.ndarray, Bd: np.ndarray):
        self.Ad = Ad
        self.Bd = Bd
        self.nx, self.nu = Bd.shape

    def update_model_residual(self, cd: np.ndarray):
        if cd is not None:
            self.cd = cd
        else:
            self.cd = np.zeros(self.nx)

    def update_horizon(self, N: int):
        self.N = N

    def update_ddl(self, ddl: Union[int, None]):
        if ddl is None:
            if hasattr(self, 'ddl'):
                delattr(self, 'ddl')
        else:
            self.ddl = ddl
        assert not hasattr(self, 'ddl') or self.ddl <= self.N

    def update_target_set(self, target_lo: np.ndarray, target_up: np.ndarray):
        self.xtmin = target_lo
        self.xtmax = target_up

    def update_safe_set(self, safe_lo: np.ndarray, safe_up: np.ndarray):
        self.xmin = safe_lo
        self.xmax = safe_up

    def formulate(self):
        """
           x = | x0, x1, ..., xN, u0, u1, ..., u{N-1} |
           Aeq                                     leq (1-D)
            | -I                              |     | -x0 |
            | Ad -I              Bd           |     |     |
            |    Ad -I              Bd        |     |     |
            |        ...             ...      |     |     |
            |           Ad -I            Bd   |     |     |
        """
        if not self.ready_to_formulate():
            raise ValueError('Not ready to formulate')
        Ax = np.kron(np.eye(self.N + 1), -np.eye(self.nx)) + np.kron(np.eye(self.N + 1, k=-1), self.Ad)
        Bu = np.kron(np.vstack([np.zeros((1, self.N)), np.eye(self.N)]), self.Bd)
        self.Aeq = np.hstack([Ax, Bu])
        self.Aineq = np.eye((self.N + 1) * self.nx + self.N * self.nu)

        repeated_cd = np.array(self.cd.tolist() * self.N)
        self.leq = np.hstack([-self.x0, -1*repeated_cd]) # self.leq = np.hstack([-self.x0, np.zeros(self.N * self.nx)])
        if hasattr(self, 'ddl') and self.ddl <= self.N:
            self.lineq = np.hstack(
                [np.kron(np.ones(self.ddl), self.xmin), np.kron(np.ones(self.N - self.ddl + 1), self.xtmin),
                 np.kron(np.ones(self.N), self.umin)])
            self.uineq = np.hstack(
                [np.kron(np.ones(self.ddl), self.xmax), np.kron(np.ones(self.N - self.ddl + 1), self.xtmax),
                 np.kron(np.ones(self.N), self.umax)])
        else:
            self.lineq = np.hstack([np.kron(np.ones(self.N + 1), self.xmin), np.kron(np.ones(self.N), self.umin)])
            self.uineq = np.hstack([np.kron(np.ones(self.N + 1), self.xmax), np.kron(np.ones(self.N), self.umax)])

        self.x = cp.Variable((self.N + 1) * self.nx + self.N * self.nu)
        self.prob = cp.Problem(cp.Minimize(0),
                               [self.Aineq @ self.x <= self.uineq,
                                self.lineq <= self.Aineq @ self.x,
                                self.Aeq @ self.x == self.leq])

    def formulate_only_x0(self):
        self.leq[:self.nx] = -self.x0
        # self.prob = cp.Problem(self.prob.objective,
        #                        [self.Aineq @ self.x <= self.uineq,
        #                         self.lineq <= self.Aineq @ self.x,
        #                         self.Aeq @ self.x == self.leq])

    def update(self, feedback_value: np.ndarray, current_time=None):
        self.x0 = feedback_value
        if not hasattr(self, 'prob'):
            self.formulate()
        else:
            self.formulate_only_x0()
        self.prob.solve(solver=self.solver, max_iter=10000, warm_start=True)

        if self.prob.status != cp.OPTIMAL:
            print('did not get an optimal solution!')
            raise ValueError('did not solve the problem!')
        # Apply first utils input to the plant
        # res.x = [ x0, x1, ... xN, u0, u1, u{N-1} ]
        ctrl = self.x.value[-self.N * self.nu:-(self.N - 1) * self.nu]
        return ctrl

    def get_full_ctrl(self):
        """
        Only call after self.update to get full utils input sequence  u0,...,u{N-1}
        """
        return self.x.value[-self.N * self.nu:].reshape((-1, self.nu))

    def get_last_x(self):
        """
        Only call after self.update to get last system state xN for debug
        """
        return self.x.value[self.N * self.nx:(self.N+1) * self.nx]

    def set_control_limit(self, control_lo: np.ndarray, control_up: np.ndarray):
        self.umin = control_lo
        self.umax = control_up

    def set_reference(self, ref: np.ndarray):
        self.xr = ref

    def set_solver(self, solver):
        self.solver = solver


import numpy as np
from scipy.linalg import solve_continuous_are, inv

from .controller_base import Controller


class LQR(Controller):
    def __init__(self, A: np.ndarray, B: np.ndarray, Q: np.ndarray, R: np.ndarray):
        self.update_gain(A, B, Q, R)
        self.ref = 0
        self.control_lo = None
        self.control_up = None

    def update_gain(self, A: np.ndarray, B: np.ndarray, Q: np.ndarray, R: np.ndarray):
        P = solve_continuous_are(A, B, Q, R)
        self.K = inv(R) @ (B.T @ P)

    def update(self, feedback_value: np.ndarray, current_time=None) -> np.ndarray:
        cin = -self.K @ (feedback_value - self.ref)
        if self.control_lo and self.control_up:
            for i in range(len(cin)):
                cin[i] = np.clip(cin[i], self.control_lo, self.control_up)
        return cin

    def set_control_limit(self, control_lo: np.ndarray, control_up: np.ndarray):
        self.control_lo = control_lo
        self.control_up = control_up

    def set_reference(self, ref: np.ndarray):
        self.ref = ref



# Ref: https://ctms.engin.umich.edu/CTMS/index.php?example=MotorSpeed&section=ControlPID

import numpy as np

from cpsim import Simulator
from cpsim.controllers.PID import PID

# system dynamics
J = 0.01   # moment of inertia of the rotor
b = 0.1    # motor viscous friction constant
Ke = 0.01  # electromotive force constant
Kt = 0.01  # motor torque constant
R = 1      # electric resistance
L = 0.5    # electric inductance

A = [[-b / J, Kt / J], [-Ke / L, -R / L]]
B = [[0], [1 / L]]
C = [[1, 0]]

x_0 = np.array([0.0, 0.0])

# utils parameters
P = 19
I = 37
D = 0.1

control_limit = {
    'lo': np.array([0]),
    'up': np.array([60])
}


class Controller:
    def __init__(self, dt):
        self.dt = dt
        self.pid = PID(P, I, D, current_time=-dt)
        self.pid.setWindup(100)
        self.pid.setSampleTime(dt)
        self.set_control_limit(control_limit['lo'], control_limit['up'])

    def update(self, ref: np.ndarray, feedback_value: np.ndarray, current_time) -> np.ndarray:
        self.pid.set_reference(ref[0])
        cin = self.pid.update(feedback_value[0], current_time)
        return np.array([cin])

    def set_control_limit(self, control_lo, control_up):
        self.control_lo = control_lo
        self.control_up = control_up
        self.pid.set_control_limit(self.control_lo[0], self.control_up[0])

    def clear(self):
        self.pid.clear(current_time=-self.dt)


class MotorSpeed(Simulator):
    """
            States: (2,)
                x[0]: the rotational speed of the shaft
                x[1]: electric current
            Control Input: (1,)
                u[0]: voltage source
            Output:  (1,)
                y[0]: the rotational speed of the shaft
                Output Feedback
            Controller: PID
            """
    def __init__(self, name, dt, max_index, noise=None):
        super().__init__('Motor Speed ' + name, dt, max_index)
        self.linear(A, B, C)
        controller = Controller(dt)
        settings = {
            'init_state': x_0,
            'feedback_type': 'output',
            'controller': controller
        }
        if noise:
            settings['noise'] = noise
        self.sim_init(settings)


if __name__ == "__main__":
    max_index = 500
    dt = 0.02
    ref = [np.array([5])] * 201 + [np.array([4])] * 200 + [np.array([5])] * 100
    noise = {
        'process': {
            'type': 'white',
            'param': {'C': np.eye(2) * 0.1}
        }
    }
    motor_speed = MotorSpeed('test', dt, max_index, noise)
    for i in range(0, max_index + 1):
        assert motor_speed.cur_index == i
        motor_speed.update_current_ref(ref[i])
        # attack here
        motor_speed.evolve()
    # print results
    import matplotlib.pyplot as plt

    t_arr = np.linspace(0, 10, max_index + 1)
    ref = [x[0] for x in motor_speed.refs[:max_index + 1]]
    y_arr = [x[0] for x in motor_speed.outputs[:max_index + 1]]

    plt.plot(t_arr, y_arr, t_arr, ref)
    plt.show()

    u_arr = [x[0] for x in motor_speed.inputs[:max_index + 1]]
    plt.plot(t_arr, u_arr)
    plt.show()



import numpy as np

from .interval_helpers import *


class NonlinearEstimator:
    def __init__(self, ode, dt):
        self.ode = ode
        self.dt = dt

    def estimate(self, x_a, us, xs, unsafe_states_onehot):
        x_a, us = x_a.tolist(), us.tolist()
        if xs is not None:
            xs = xs.tolist()            
        x_a_interval = convert_1D_list_to_intervals(x_a)
        # print(f'inside estimate.... \n')
        # print(f'{x_a_interval=}, {us=}')
        x_0_interval = reach(self.ode, x_a_interval, us, xs, unsafe_states_onehot, step_size=self.dt)
        # print(f'{x_0_interval=}, {us=}')
        x_0_lo = [Int[0][0] for Int in x_0_interval]
        x_0_up = [Int[0][1] for Int in x_0_interval]
        x_0 = [midpoint(Int) for Int in x_0_interval]
        return np.array(x_0_lo), np.array(x_0_up), np.array(x_0)

    def get_deadline(self, x_a, safe_set_lo, safe_set_up, control: np.array, max_k, verbose=True):
        k = max_k
        breaked = False
        for i in range(max_k):
            i += 1
            control_series = np.array([control] * i)
            state_series = None # Not using safe states for deadline
            unsafe_states_onehot = [1] * len(x_a)
            x_0_lo, x_0_up, x_0 = self.estimate(x_a, control_series, state_series, unsafe_states_onehot)
            if verbose:
                print(f'within deadline computation... @ {i=}/{max_k}, {x_0_lo=} and {x_0_up=} and {x_0=}\n')
            for j in range(np.size(x_a, 0)):
                if safe_set_lo[j] < x_0_lo[j] < safe_set_up[j] and safe_set_lo[j] < x_0_up[j] < safe_set_up[j]:
                    pass
                else:
                    breaked = True
                    break
            if breaked:
                k = i - 1
                break
        return k



import numpy as np
from interval import imath
from numpy import exp

from cpsim import Simulator
from cpsim.controllers.PID_incremental import PID

# Parameters:
# Volumetric Flowrate (m^3/sec)
q = 100
# Volume of CSTR (m^3)
V = 100
# Density of A-B Mixture (kg/m^3)
rho = 1000
# Heat capacity of A-B Mixture (J/kg-K)
Cp = 0.239
# Heat of reaction for A->B (J/mol)
mdelH = 5e4
# E - Activation energy in the Arrhenius Equation (J/mol)
# R - Universal Gas Constant = 8.31451 J/mol-K
EoverR = 8750
# Pre-exponential factor (1/sec)
k0 = 7.2e10
# U - Overall Heat Transfer Coefficient (W/m^2-K)
# A - Area - this value is specific for the U calculation (m^2)
UA = 5e4


def cstr(t, x, u, Tf=350, Caf=1, use_imath=False):
    # Inputs (3):
    # Temperature of cooling jacket (K)
    Tc = u[0]
    # Tf = Feed Temperature (K)
    # Caf = Feed Concentration (mol/m^3)

    # States (2):
    # Concentration of A in CSTR (mol/m^3)
    Ca = x[0]
    # Temperature in CSTR (K)
    T = x[1]

    # reaction rate
    if use_imath:
        rA = k0 * imath.exp(-EoverR / T) * Ca
    else:
        rA = k0 * np.exp(-EoverR / T) * Ca

    # Calculate concentration derivative
    dCadt = q / V * (Caf - Ca) - rA
    # Calculate temperature derivative
    dTdt = q / V * (Tf - T) \
           + mdelH / (rho * Cp) * rA \
           + UA / V / rho / Cp * (Tc - T)

    # Return xdot:
    if use_imath:
        return [dCadt, dTdt]
    else:
        xdot = np.zeros(2)
        xdot[0] = dCadt
        xdot[1] = dTdt
        return xdot

def cstr_imath(t, x, u, Tf=350, Caf=1):
    return cstr(t=t, x=x, u=u, Tf=Tf, Caf=Caf, use_imath=True)


def f(x, u, dt):
    dx = cstr(None, x, u)
    return x + dt * dx

def jfx(x, u, dt):
    Ca = x[0]
    T = x[1]
    Tc = u[0]
    Ad = np.array([
        [dt * (-k0 * exp(-EoverR / T) - q / V) + 1, -Ca * EoverR * dt * k0 * exp(-EoverR / T) / T ** 2],
        [dt * k0 * mdelH * exp(-EoverR / T) / (Cp * rho),
         dt * (Ca * EoverR * k0 * mdelH * exp(-EoverR / T) / (Cp * T ** 2 * rho) - q / V - UA / (Cp * V * rho)) + 1]])
    return Ad

def jfu(x, u, dt):
    Ca = x[0]
    T = x[1]
    Tc = u[0]
    Bd = np.array([
        [0],
        [UA * dt / (Cp * V * rho)]])
    return Bd


# initial states
x_0 = np.array([0.98, 280])

# utils parameters
KP = 0.5 * 1.0
KI =  KP / (3 / 8.0)
KD =  - KP * 0.1
u_0 = 274.57786
control_limit = {
    'lo': np.array([250]),
    'up': np.array([350])
}


class Controller:
    def __init__(self, dt):
        self.dt = dt
        self.pid = PID(u_0, KP, KI, KD, current_time=-dt)
        self.pid.setWindup(100)
        self.pid.setSampleTime(dt)
        self.set_control_limit(control_limit['lo'], control_limit['up'])

    def update(self, ref: np.ndarray, feedback_value: np.ndarray, current_time) -> np.ndarray:
        self.pid.set_reference(ref[1])             # only care about the 2nd state here
        cin = self.pid.update(feedback_value[1], current_time)  # only use the 2nd state here
        return np.array([cin])

    def set_control_limit(self, control_lo, control_up):
        self.control_lo = control_lo
        self.control_up = control_up
        self.pid.set_control_limit(self.control_lo[0], self.control_up[0])

    def clear(self):
        self.pid.clear(current_time=-self.dt)


class CSTR(Simulator):
    """
        States: (2,)
            x[0]: Concentration of A in CSTR (mol/m^3)
            x[1]: Temperature in CSTR (K)
        Control Input: (1,)
            u[0]: Temperature of cooling jacket (K)
        Output:  (2,)
            State Feedback
        Controller: PID
    """

    def __init__(self, name, dt, max_index, noise=None):
        super().__init__('CSTR ' + name, dt, max_index)
        self.nonlinear(ode=cstr, n=2, m=1, p=2)  # number of states, utils inputs, outputs
        controller = Controller(dt)
        settings = {
            'init_state': x_0,
            'feedback_type': 'state',  # 'state' or 'output',  you must define C if 'output'
            'controller': controller
        }
        if noise:
            settings['noise'] = noise
        self.sim_init(settings)


if __name__ == "__main__":
    max_index = 1000
    dt = 0.02
    ref = [np.array([0, 300])] * (max_index+1)
    # noise = {
    #     'process': {
    #         'type': 'box_uniform',
    #         'param': {'lo': np.array([-0.000001, -0.001]), 'up': np.array([0.001, 0.001])}
    #     }
    # }
    noise = None
    cstr_model = CSTR('test', dt, max_index, noise)


    for i in range(0, max_index + 1):
        assert cstr_model.cur_index == i
        cstr_model.update_current_ref(ref[i])
        # attack here

        # x = A x + B u
        # x, u - 1-D array
        # cstr_model.cur_x   #  ground truth
        # cstr_model.cur_u   #  utils input of last step
        # cstr_model.cur_y   #
        # cstr_model.cur_feedback   # attack on this

        cstr_model.evolve()


    # print results
    import matplotlib.pyplot as plt

    t_arr = np.linspace(0, 10, max_index + 1)
    ref = [x[1] for x in cstr_model.refs[:max_index + 1]]
    y_arr = [x[1] for x in cstr_model.outputs[:max_index + 1]]

    plt.plot(t_arr, y_arr, t_arr, ref)
    plt.show()

    u_arr = [x[0] for x in cstr_model.inputs[:max_index + 1]]
    plt.plot(t_arr, u_arr)
    plt.show()



from time import perf_counter

class Timer:
    def __init__(self):
        self.total_time = 0
        self.tic()

    def tic(self):
        self.start = perf_counter()

    def toc(self):
        self.total_time += perf_counter() - self.start

    def reset(self):
        self.total_time = 0
        self.tic()

    def total(self):
        return self.total_time



import numpy as np
from scipy.linalg import solve_continuous_are, inv

from .controller_base import Controller


class LQRSSE(Controller):
    def __init__(self, A: np.ndarray, B: np.ndarray, Q: np.ndarray, R: np.ndarray):
        self.update_gain(A, B, Q, R)
        self.ref = 0
        self.control_lo = None
        self.control_up = None

    def update_gain(self, A: np.ndarray, B: np.ndarray, Q: np.ndarray, R: np.ndarray):
        P = solve_continuous_are(A, B, Q, R)
        self.K = inv(R) @ (B.T @ P)

    def update(self, feedback_value: np.ndarray, current_time=None) -> np.ndarray:
        # cin = -self.K @ (feedback_value - self.ref)
        # if self.control_lo and self.control_up:
        #     for i in range(len(cin)):
        #         cin[i] = np.clip(cin[i], self.control_lo[i], self.control_up[i])
        cin = -self.K @ (feedback_value - self.ref * self.Nx )
        cin = cin + feedback_value * self.Nu
        if self.control_lo and self.control_up:
            for i in range(len(cin)):
                cin[i] = np.clip(cin[i], self.control_lo[i], self.control_up[i])
        return cin

    def set_control_limit(self, control_lo: np.ndarray, control_up: np.ndarray):
        self.control_lo = control_lo
        self.control_up = control_up

    def set_reference(self, ref: np.ndarray):
        self.ref = ref

    def compute_Nbar(self, A:np.ndarray, B:np.ndarray, C:np.ndarray, D:np.ndarray):
        s = A.shape[0]
        Z = np.zeros([1, s])
        Z = np.concatenate((Z, np.array([[1]])), axis=1)
        con = np.concatenate((A, B), axis=1)
        con2 = np.concatenate((C, D), axis=1)
        temp = np.concatenate((con, con2), axis=0)
        N = inv(temp) @ np.transpose(Z)
        self.Nx = N[0:s]
        self.Nu = N[s]
        self.Nbar=self.Nu + self.K*self.Nx




# Ref: Data-Driven Science and Engineering: Machine Learning, Dynamical Systems, and Control (Session 8.7, Page 300)

import math

import numpy as np

from cpsim import Simulator
from cpsim.controllers.PID import PID


def quad(t, x, u, use_imath=False):
    # Control Inputs (4):
    # Only U_t is non-zero

    # More Parameters:
    Ix = 5 * 0.001
    Iy = 5 * 0.001
    Iz = 5 * 0.001 # values from https://repository.upenn.edu/cgi/viewcontent.cgi?article=1705&context=edissertations
    mass = 1
    g = 9.81

    # States (12):
    phi, theta, psi, w_phi, w_theta, w_psi, x, y, z, v_x, v_y, v_z = x
    U_t = u[0]
    U_phi = 0
    U_theta = 0
    U_psi  = 0
    if use_imath:
        from interval import imath
        xdot = [
        w_phi,
        w_theta,
        w_psi,
        U_phi/Ix + w_theta * w_psi * (Iy - Iz)/Ix, # 0
        U_theta/Iy + w_phi * w_psi * (Iz - Iz)/Iy, # 0
        U_psi/Iz + w_phi * w_theta * (Ix - Iy)/Iz, # 0
        v_x,
        v_y,
        v_z,
        # Replaced theta with phi on RHS terms below based on https://arxiv.org/pdf/1602.02622.pdf
        U_t/mass * (imath.cos(phi)* imath.sin(theta) * imath.cos(psi) + imath.sin(phi) * imath.sin(psi)),
        U_t/mass * (imath.cos(phi)* imath.sin(theta) * imath.sin(psi) - imath.sin(phi) * imath.cos(psi)),
        U_t/mass * imath.cos(phi) * imath.cos(theta) - g
    ]
    else:
        xdot = np.array([
        w_phi,
        w_theta,
        w_psi,
        U_phi/Ix + w_theta * w_psi * (Iy - Iz)/Ix, # 0
        U_theta/Iy + w_phi * w_psi * (Iz - Iz)/Iy, # 0
        U_psi/Iz + w_phi * w_theta * (Ix - Iy)/Iz, # 0
        v_x,
        v_y,
        v_z,
        # Replaced theta with phi on RHS terms below based on https://arxiv.org/pdf/1602.02622.pdf
        U_t/mass * (math.cos(phi)* math.sin(theta) * math.cos(psi) + math.sin(phi) * math.sin(psi)),
        U_t/mass * (math.cos(phi)* math.sin(theta) * math.sin(psi) - math.sin(phi) * math.cos(psi)),
        U_t/mass * math.cos(phi) * math.cos(theta) - g
    ])
    # xdot = [
    #     0,
    #     0,
    #     0,
    #     0, # 0
    #     0, # 0
    #     0, # 0
    #     0,
    #     0,
    #     v_z,
    #     # Replaced theta with phi on RHS terms below based on https://arxiv.org/pdf/1602.02622.pdf
    #     0,
    #     0,
    #     U_t/mass - g
    # ]
    # Return xdot:
    return xdot
    
def quad_imath(t, x, u):
    return quad(t=t, x=x, u=u, use_imath=True)

def quad_jfx(x, u, dt):
    # More Parameters:
    Ix = 5 * 0.001
    Iy = 5 * 0.001
    Iz = 5 * 0.001 # values from https://repository.upenn.edu/cgi/viewcontent.cgi?article=1705&context=edissertations
    mass = 1
    g = 9.81

    # States (12):
    phi, theta, psi, w_phi, w_theta, w_psi, x, y, z, v_x, v_y, v_z = x
    U_t = u[0]
    U_phi = 0
    U_theta = 0
    U_psi  = 0

    Jacobian = np.array([
    [                                                              1,                                        0,                                                              0,                      dt,                     0,                       0, 0, 0, 0,  0,  0,  0],
    [                                                              0,                                        1,                                                              0,                       0,                    dt,                       0, 0, 0, 0,  0,  0,  0],
    [                                                              0,                                        0,                                                              1,                       0,                     0,                      dt, 0, 0, 0,  0,  0,  0],
    [                                                              0,                                        0,                                                              0,                       1, dt*w_psi*(Iy - Iz)/Ix, dt*w_theta*(Iy - Iz)/Ix, 0, 0, 0,  0,  0,  0],
    [                                                              0,                                        0,                                                              0,                       0,                     1,                       0, 0, 0, 0,  0,  0,  0],
    [                                                              0,                                        0,                                                              0, dt*w_theta*(Ix - Iy)/Iz, dt*w_phi*(Ix - Iy)/Iz,                       1, 0, 0, 0,  0,  0,  0],
    [                                                              0,                                        0,                                                              0,                       0,                     0,                       0, 1, 0, 0, dt,  0,  0],
    [                                                              0,                                        0,                                                              0,                       0,                     0,                       0, 0, 1, 0,  0, dt,  0],
    [                                                              0,                                        0,                                                              0,                       0,                     0,                       0, 0, 0, 1,  0,  0, dt],
    [U_t*dt*(-math.sin(phi)*math.sin(theta)*math.cos(psi) + math.sin(psi)*math.cos(phi))/mass, U_t*dt*math.cos(phi)*math.cos(psi)*math.cos(theta)/mass, U_t*dt*(math.sin(phi)*math.cos(psi) - math.sin(psi)*math.sin(theta)*math.cos(phi))/mass,                       0,                     0,                       0, 0, 0, 0,  1,  0,  0],
    [U_t*dt*(-math.sin(phi)*math.sin(psi)*math.sin(theta) - math.cos(phi)*math.cos(psi))/mass, U_t*dt*math.sin(psi)*math.cos(phi)*math.cos(theta)/mass, U_t*dt*(math.sin(phi)*math.sin(psi) + math.sin(theta)*math.cos(phi)*math.cos(psi))/mass,                       0,                     0,                       0, 0, 0, 0,  0,  1,  0],
    [                               -U_t*dt*math.sin(phi)*math.cos(theta)/mass,         -U_t*dt*math.sin(theta)*math.cos(phi)/mass,                                                              0,                       0,                     0,                       0, 0, 0, 0,  0,  0,  1]
    ])

    return Jacobian

x_0 = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
control_limit = {
    'lo': np.array([-10]),
    'up': np.array([100])
}

# utils parameters


KP =  100
KI = 0
KD = -19
class Controller:
    def __init__(self, dt):
        self.dt = dt
        self.pid = PID(KP, KI, KD, current_time=-dt)
        self.pid.setWindup(0)
        self.pid.setSampleTime(dt)
        self.set_control_limit(control_limit['lo'], control_limit['up'])

    def update(self, ref: np.ndarray, feedback_value: np.ndarray, current_time) -> np.ndarray:
        self.pid.set_reference(ref[-4])
        cin = self.pid.update(feedback_value[-4], current_time)
        return np.array([cin])

    def set_control_limit(self, control_lo, control_up):
        self.control_lo = control_lo
        self.control_up = control_up
        self.pid.set_control_limit(self.control_lo[0], self.control_up[0])

    def clear(self):
        self.pid.clear(current_time=-self.dt)



class quadrotor(Simulator):
    """
    States: (4,)
        x[0]: location of cart
        x[1]: dx[0]
        x[2]: pendulum angle  (down:0, up:pi)
        x[3]: dx[1]
    Control Input: (1,)  [control_limit]
        u[0]: force on the cart
    Output: (4,)
        State Feedback
    Controller: LQR
    """

    def __init__(self, name, dt, max_index, noise=None):
        super().__init__('Quadrotor ' + name, dt, max_index)
        self.nonlinear(ode=quad, n=12, m=1, p=12)
        controller = Controller(dt)
        settings = {
            'init_state': x_0,
            'feedback_type': 'state',
            'controller': controller
        }
        if noise:
            settings['noise'] = noise
        self.sim_init(settings)


if __name__ == "__main__":
    max_index = 300
    dt = 0.01
    ref = [np.array([0, 0, 0, 0, 0, 0, 0, 0, 10, 0, 0, 0])] * (max_index+1)
    # bias attack example
    from cpsim import Attack
    bias = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
    bias_attack = Attack('bias', bias, 300)
    ip = quadrotor('test', dt, max_index)
    for i in range(0, max_index + 1):
        assert ip.cur_index == i
        ip.update_current_ref(ref[i])
        # attack here
        # ip.cur_feedback = bias_attack.launch(ip.cur_feedback, ip.cur_index, ip.states)
        ip.evolve()
    # print results
    import matplotlib.pyplot as plt

    fig, ax = plt.subplots(2, 1)
    ax1, ax2 = ax
    t_arr = np.linspace(0, 3, max_index + 1)
    ref2 = [x[-4] for x in ip.refs[:max_index + 1]]
    y2_arr = [x[-4] for x in ip.outputs[:max_index + 1]]
    ax1.set_title('Altitude')
    ax1.plot(t_arr, y2_arr, t_arr, ref2)
    u_arr = [x[0] for x in ip.inputs[:max_index + 1]]
    ax2.set_title('u-force')
    ax2.plot(t_arr, u_arr)
    plt.show()



import numpy as np


class Attack:
    def  __init__(self, category, param, start_index, end_index=None):
        """
        category: 'bias', 'delay', 'replay'  todo: 'stealthy'
        param
        """
        self.cat = category
        self.param = param
        self.start_index = start_index
        self.end_index = end_index if end_index is not None else np.Inf
        self.cur_index = None
        self.disabled = False

    def update_param(self, param):
        self.param = param

    def disable(self):
        self.disabled = True

    def launch(self, cur_data, cur_index, history_intact_data):
        self.cur_index = cur_index
        if cur_index < self.start_index or cur_index > self.end_index or self.disabled:
            return cur_data
        if self.cat == 'bias':
            # param is the bias on each element
            return cur_data + self.param
        if self.cat == 'delay':
            # param is the number of delay steps
            if self.cur_index - self.start_index > self.param:
                return history_intact_data[self.cur_index - self.param]
            else:
                return history_intact_data[self.start_index]
        if self.cat == 'replay':
            # replay interval param['start'] ~ param['end']
            delta_index = self.cur_index - self.start_index
            offset_index = self.param['start'] + delta_index % (self.param['end']-self.param['start'])
            return history_intact_data[offset_index]

    def get_C_filter(self):
        if self.cat == 'bias':
            cnt = len(self.param)
            bias_e = np.full((cnt,), False)
            for i in range(cnt):
                if self.param[i] != 0:
                    bias_e[i] = True
            C_filter = []
            for i in range(4):
                if bias_e[i]:
                    continue
                else:
                    tmp = [0, 0, 0, 0]
                    tmp[i] = 1
                    C_filter.append(tmp)
            return C_filter
        else:
            raise NotImplementedError()


if __name__ == "__main__":
    attack = Attack('bias', np.array([1, 0, 0, 0]), 0)
    print(attack.get_C_filter())



from copy import deepcopy

import sympy as sym

# ----------------------------------------------------
Ix, Iy, Iz, mass, g, phi, theta, psi, w_phi, w_theta, w_psi, x, y, z, v_x, v_y, v_z, U_t, U_phi, U_theta, U_psi  = sym.symbols('Ix Iy Iz mass g phi theta psi w_phi w_theta w_psi x y z v_x v_y v_z U_t U_phi U_theta U_psi')

vars = [
    phi,
    theta,
    psi,
    w_phi,
    w_theta,
    w_psi,
    x,
    y,
    z,
    v_x,
    v_y,
    v_z,
]

us = [
    U_phi,
    U_theta,
    U_psi,
    U_t,
]

f = [
    w_phi, 
    w_theta, 
    w_psi, 
    U_phi/Ix + w_theta * w_psi * (Iy - Iz)/Ix, 
    U_theta/Iy + w_phi * w_psi * (Iz - Iz)/Iy, 
    U_psi/Iz + w_phi * w_theta * (Ix - Iy)/Iz, 
    v_x,
    v_y,
    v_z,
    U_t/mass * (sym.cos(phi)* sym.sin(theta) * sym.cos(psi) + sym.sin(phi) * sym.sin(psi)),
    U_t/mass * (sym.cos(phi)* sym.sin(theta) * sym.sin(psi) - sym.sin(phi) * sym.cos(psi)),
    U_t/mass * sym.cos(phi) * sym.cos(theta) - g,
]
# ----------------------------------------------------

dt = sym.symbols('dt')
J = sym.zeros(len(f), len(vars))
for i, fi in enumerate(f):
    for j, s in enumerate(vars):
        J[i, j] = sym.diff(fi, s)
Ac = deepcopy(J)
Ad = sym.eye(len(vars)) + dt * Ac
print(f'{Ad=}')

# J = sym.zeros(len(f), len(us))
# for i, fi in enumerate(f):
#     for j, s in enumerate(us):
#         J[i, j] = sym.diff(fi, s)
# Bc = deepcopy(J)
# Bd = dt * Bc
# print(f'{Bd=}')







from copy import deepcopy

import sympy as sym

# ----------  modify -----------------------------
q, V, rho, Cp, mdelH, EoverR, k0, UA, Tc, Ca, T, Caf, Tf  = sym.symbols('q V rho Cp mdelH EoverR k0 UA Tc Ca T Caf Tf')

rA = k0 * sym.exp(-EoverR / T) * Ca
# Calculate concentration derivative
dCadt = q / V * (Caf - Ca) - rA
# Calculate temperature derivative
dTdt = q / V * (Tf - T) \
       + mdelH / (rho * Cp) * rA \
       + UA / V / rho / Cp * (Tc - T)

vars = [Ca, T]
us = [Tc]
f = [dCadt, dTdt]
# ----------------------------------------------------

dt = sym.symbols('dt')
J = sym.zeros(len(f), len(vars))
for i, fi in enumerate(f):
    for j, s in enumerate(vars):
        J[i, j] = sym.diff(fi, s)
Ac = deepcopy(J)
Ad = sym.eye(len(vars)) + dt * Ac
print(f'{Ad=}')

J = sym.zeros(len(f), len(us))
for i, fi in enumerate(f):
    for j, s in enumerate(us):
        J[i, j] = sym.diff(fi, s)
Bc = deepcopy(J)
Bd = dt * Bc
print(f'{Bd=}')







from abc import ABC, abstractmethod


class Controller(ABC):
    @abstractmethod
    def update(self, feedback_value, current_time=None):
        pass

    @abstractmethod
    def set_control_limit(self, control_lo, control_up):
        pass

    @abstractmethod
    def set_reference(self, ref):
        pass

    def clear(self):
        pass



#Ref:https://www.kth.se/polopoly_fs/1.588039.1600688317!/Thesis%20KTH%20-%20Francesco%20Sabatino.pdf
import numpy as np

from cpsim import Simulator
from cpsim.controllers.PID import PID

# system dynamics
g = 9.81
m = 0.468
A = [[0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0],
     [0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
     [0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
     [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
     [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
     [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
     [0, -g, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
     [g, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
     [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
     [0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
     [0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
     [0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0]]
B = [[0], [0], [0], [0], [0], [0], [0], [0], [1 / m], [0], [0], [0]]
C = [[1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
     [0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
     [0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
     [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
     [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
     [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1]]
D = [[0], [0], [0], [0], [0], [0]]

x_0 = np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])

# utils parameters
control_limit = {
    'lo': np.array([-50]),
    'up': np.array([50])
}

KP = 15
KI = 0
KD = -4


class Controller:
    def __init__(self, dt):
        self.dt = dt
        self.pid = PID(KP, KI, KD, current_time=-dt)
        self.pid.setWindup(100)
        self.pid.setSampleTime(dt)
        self.set_control_limit(control_limit['lo'], control_limit['up'])

    def update(self, ref: np.ndarray, feedback_value: np.ndarray, current_time) -> np.ndarray:
        self.pid.set_reference(ref[0])
        cin = self.pid.update(feedback_value[5], current_time)
        return np.array([cin])

    def set_control_limit(self, control_lo, control_up):
        self.control_lo = control_lo
        self.control_up = control_up
        self.pid.set_control_limit(self.control_lo[0], self.control_up[0])

    def clear(self):
        self.pid.clear(current_time=-self.dt)


class Quadrotor(Simulator):
    """
               States: (12,)
                    x[0], x[1], x[2]: Roll, Pitch, Yaw
                    x[9], x[10], x[11]: relative position
                    x[11]: altitude
               Control Input: (4,)
                   u[0]: vertical thrust
                   u[1], u[2], u[3]:  one for each of the angular motions
               Output:  (6,)
                   y[5]: the altitude of the quadrotor
                   Output Feedback
               Controller: PID
               """
    def __init__(self, name, dt, max_index, noise=None):
        super().__init__('Quadrotor ' + name, dt, max_index)
        self.linear(A, B, C)
        controller = Controller(dt)
        settings = {
            'init_state': x_0,
            'feedback_type': 'output',
            'controller': controller
        }
        if noise:
            settings['noise'] = noise
        self.sim_init(settings)


if __name__ == "__main__":
    max_index = 600
    dt = 0.02
    ref = [np.array([2])] * 201 + [np.array([4])] * 200 + [np.array([2])] * 200
    noise = {
        'process': {
            'type': 'white',
            'param': {'C': np.eye(12) * 0.001}
        }
    }
    quadrotor = Quadrotor('test', dt, max_index, None)
    for i in range(0, max_index + 1):
        assert quadrotor.cur_index == i
        quadrotor.update_current_ref(ref[i])
        # attack here
        quadrotor.evolve()
    # print results
    import matplotlib.pyplot as plt

    t_arr = np.linspace(0, 10, max_index + 1)
    ref = [x[0] for x in quadrotor.refs[:max_index + 1]]
    y_arr = [x[5] for x in quadrotor.outputs[:max_index + 1]]

    plt.plot(t_arr, y_arr, t_arr, ref)
    plt.show()



import numpy as np
from scipy.linalg import inv

class KalmanFilter:
    def __init__(self, Ad: np.ndarray, Bd: np.ndarray, C: np.ndarray, D: [np.ndarray, None], Q: np.ndarray, R: np.ndarray):
        self.A = Ad
        self.B = Bd
        self.C = C
        self.D = D if D is None else np.zeros((self.C.shape[0], self.B.shape[1]))
        self.Q = Q
        self.R = R

    def predict(self, x: np.ndarray, P: np.ndarray, u: np.ndarray):
        x_predict = self.A @ x + self.B @ u
        P_predict = self.A @ P @ self.A.T + self.Q
        return x_predict, P_predict

    def update(self, x_predict, P_predict, y: np.ndarray):
        K = P_predict @ self.C.T @ inv(self.C @ P_predict @ self.C.T + self.R)
        x_update = x_predict + K @ (y-self.C @ x_predict)
        P_update = P_predict - K @ self.C @ P_predict
        return x_update, P_update

    def one_step(self, x: np.ndarray, P: np.ndarray, u: np.ndarray, y: np.ndarray):
        x_predict, P_predict = self.predict(x, P, u)
        x_update, P_update = self.update(x_predict, P_predict, y)
        return x_update, P_update

    def multi_steps(self, x: np.ndarray, P: np.ndarray, us: np.ndarray, ys: np.ndarray):
        """
        us: shape (length, m)
        ys: shape (length, p)
        """
        assert len(us) == len(ys)
        length = len(us)
        x_res = np.empty((length+1, *x.shape))
        P_res = np.empty((length+1, *P.shape))
        x_res[0] = x
        P_res[0] = P
        for i in range(length):
            x_res[i+1], P_res[i+1] = self.one_step(x_res[i], P_res[i], us[i], ys[i])
        return x_res, P_res




# for plot
from itertools import product

import matplotlib.pyplot as plt
import numpy as np
from scipy.spatial import ConvexHull

from cpsim.formal.half_space import HalfSpace
from cpsim.formal.hyperplane import Hyperplane
from cpsim.formal.strip import Strip


class Zonotope:
    """
    Zonotope.
    parameters:
        c: center
        g: [g_0, g_1, ...]  generators
    """

    def __init__(self, c: np.ndarray, g: np.ndarray):
        self.c = c
        self.g = g
        self.dim = c.shape[0]  # dimensions
        assert self.g.shape[0] == self.dim

    def __str__(self):
        string = 'Zonotope\n  center:' + str(self.c)
        string += '\n  generators:'
        for i in range(len(self)):
            string += str(self[i])
        return string

    # '+' overload, Minkowski sum of two zonotopes
    def __add__(self, other):
        if isinstance(other, Zonotope):
            assert self.dim == other.dim
            c = self.c + other.c
            g = np.block([self.g, other.g])
            return Zonotope(c, g)
        if isinstance(other, np.ndarray):
            assert self.dim == other.shape[0]
            c = self.c + other
            return Zonotope(c, self.g)
        raise NotImplemented

    # np.array + zonotope
    def __radd__(self, other):
        return self.__add__(other)

    # '+=' overload
    def __iadd__(self, other):
        assert self.dim == other.dim
        self.c += other.c
        self.g = np.block([self.g, other.g])
        return self

    # len() overload, get the number of generators
    def __len__(self):
        return self.g.shape[1]

    # [] slicing overload, get the key-th generator
    def __getitem__(self, item):
        assert 0 <= item < self.g.shape[1]
        return self.g[:, item]

    # mask method in numpy
    __array_ufunc__ = None

    # '@' overload, linear transformation
    def __rmatmul__(self, other: np.ndarray):
        assert other.shape[1] == self.dim
        c = other @ self.c
        g = other @ self.g
        return Zonotope(c, g)

    # support function along l direction
    def support(self, l: np.ndarray):
        assert l.shape == (self.dim,)
        A_T_l = self.g.T @ l
        rho_A_T_l = np.linalg.norm(A_T_l, ord=1)
        rho_c = l @ self.c
        return rho_c + rho_A_T_l

    # all generators contributing to the farther vertex along l
    def vertex_with_max_support(self, l: np.ndarray):
        """
        Return:
            vertex - farthest vertex coordinate
            alpha  - direction (1/-1) for each generators
            gs_l   - generators to vertex
        """
        assert l.shape == (self.dim,)
        alpha = np.empty((len(self),))
        for i in range(len(self)):
            alpha[i] = -1 if self[i] @ l < 0 else 1
        gs_l = self.g @ np.diag(alpha)
        vertex = self.c + np.sum(gs_l, axis=1)
        return vertex, alpha, gs_l

    # algorithm 1
    def point_closest_to_hyperplane(self, hp: Hyperplane):
        z_star = self.c
        l = hp.l
        arrive = False
        g_num = len(self)
        alpha = np.zeros((g_num,))
        for i in range(g_num):
            alpha[i] = -1 if self[i] @ l < 0 else 1
            z_star_next = z_star+alpha[i]*self[i]
            if z_star_next @ l >= hp.b:
                alpha[i] = (hp.b-z_star@l)/(self[i]@l)
                z_star = z_star+alpha[i]*self[i]
                arrive = True
                break
            else:
                z_star = z_star_next
        return z_star, alpha, arrive

    # check if intersect with a half space
    def is_intersected(self, hs):
        if isinstance(hs, HalfSpace):
            return self.support(hs.l) >= hs.b
        if isinstance(hs, Strip):
            s1 = self.support(hs.l)
            s2 = self.support(-hs.l)
            return s1 >= hs.a and s2 <= hs.b
        if isinstance(hs, Hyperplane):
            s1 = self.support(hs.l)
            s2 = self.support(-hs.l)
            return s2 <= hs.b <= s1
        raise NotImplemented

    # get all vertices in order
    def to_V(self):
        # get all vertices
        n = self.g.shape[1]  # number of generators
        alpha = np.array(list(product(*list(zip([-1] * n, [1] * n))))).T
        v = self.c.reshape((-1, 1)) + self.g @ alpha  # all possible vertices for each column
        v = v.T  # one possible vertex per row
        v = v[ConvexHull(v).vertices, :]  # one vertex per row
        return v

    def plot(self, fig=None, color='orange'):
        v = self.to_V()
        if v.shape[1] != 2:  # only 2-d
            raise NotImplemented
        v = np.vstack((v, v[0]))  # close the polygon
        if fig is None:
            fig = plt.figure()
            plt.plot(v[:, 0], v[:, 1], color)
            plt.show()
        else:
            plt.plot(v[:, 0], v[:, 1], color)
        return fig

    # display routine by generators
    def show_routine(self, gs_l, fig=None):
        if self.dim != 2:
            raise NotImplemented
        routine = np.empty((len(self) + 1, self.dim), dtype=float)
        routine[0] = self.c
        for i in range(len(self)):
            routine[i + 1] = routine[i] + gs_l[:, i]
        if fig is None:
            fig = plt.figure()
        self.plot(fig)
        X = routine[:, 0]
        Y = routine[:, 1]
        for i in range(len(X) - 1):
            plt.arrow(X[i], Y[i], X[i + 1] - X[i], Y[i + 1] - Y[i], head_width=1.5, width=0.1,
                      length_includes_head=True, ec='g')
        if fig is None:
            plt.show()
        return fig

    # display routine by utils inputs
    def show_control_effect(self, alpha, u_dim: int, head_width, line_width, fig=None):
        if self.dim != 2:
            raise NotImplemented
        # print(len(self), u_dim)
        gs_l = self.g @ np.diag(alpha)
        routine_num = len(self) // u_dim
        routine = np.empty((routine_num + 1, self.dim), dtype=float)
        routine[0] = self.c
        for i in range(routine_num):
            # print(gs_l[:, i*u_dim:(i+1)*u_dim], np.sum(gs_l[:, i*u_dim:(i+1)*u_dim], axis=0))
            u_effect = np.sum(gs_l[:, i * u_dim:(i + 1) * u_dim], axis=1)
            routine[i + 1] = routine[i] + u_effect
        if fig is None:
            fig = plt.figure()
        self.plot(fig)
        X = routine[:, 0]
        Y = routine[:, 1]
        if head_width is None:
            head_width = 1.5
            line_width = 0.1
        for i in range(len(X) - 1):
            plt.arrow(X[i], Y[i], X[i + 1] - X[i], Y[i + 1] - Y[i], head_width=head_width, width=line_width,
                      length_includes_head=True, ec='g')
        if fig is None:
            plt.show()
        return fig

    # call from one-step utils input zonotope
    # for u, each row is a utils input
    def alpha_to_control(self, alpha):
        u_dim = len(self)
        u_num = alpha.shape[0]//u_dim
        u = np.empty((u_num, u_dim), dtype=float)
        for i in range(u_num):
            u[i] = self.g @ alpha[i*u_dim:(i+1)*u_dim] + self.c
        return u

    @classmethod
    def from_box(cls, lo: np.ndarray, up: np.ndarray):
        """
        initiate a zonotope from box
        lo: lower bound of the box
        up: upper bound of the box
        """
        c = (lo + up) / 2
        g_range = (up - lo) / 2
        g = np.diag(g_range)
        return cls(c, g)


if __name__ == '__main__':
    c1 = np.array([1, 2, 3], dtype=float)
    g1 = np.empty((3, 2))
    g1[:, 0] = np.array([1, 2, 3])
    g1[:, 1] = np.array([4, 5, 6])
    z1 = Zonotope(c1, g1)
    print("z1 =", z1)
    c2 = np.array([3, 1, 1], dtype=float)
    g2 = np.empty((3, 3))
    g2[:, 0] = np.array([9, 8, 7])
    g2[:, 1] = np.array([6, 5, 4])
    g2[:, 2] = np.array([3, 2, 1])
    z2 = Zonotope(c2, g2)
    print("z2 =", z2)

    # Minkowski sum
    z3 = z1 + z2
    print('z1+z2 = ', z3)

    # linear transformation
    A = np.array([[-1, 0, 0], [0, -1, 0], [0, 0, 1]])
    z4 = A @ z2
    print('A @ z2 =', z4)

    # support function
    c5 = np.array([1, 1], dtype=float)
    g5 = np.empty((2, 3))
    g5[:, 0] = np.array([1, 0])
    g5[:, 1] = np.array([0, -1])
    g5[:, 2] = np.array([1, 1])
    z5 = Zonotope(c5, g5)
    print('z5 =', z5)
    l = np.array([-1, -1])
    print('l=', l)
    print('support function along l:', z5.support(l))  # 2
    print('generators to farthest  vertex along l', z5.vertex_with_max_support(l))
    l = np.array([-1, 2])
    print('l=', l)
    print('support function along l:', z5.support(l))  # 5
    print('generators to farthest  vertex along l', z5.vertex_with_max_support(l))

    # check intersection
    l = np.array([-1, -1])
    bs = [4, 2, -2, -10]
    hss = [HalfSpace(l, b) for b in bs]
    res = [z5.is_intersected(hs) for hs in hss]
    print('check intersection', res)

    # plot zonotope
    v = z5.plot()

    # test alpha to utils
    alpha = np.array([1, 1, -1, 1, -1, -1])
    u_lo = np.array([-1, -4])
    u_up = np.array([3, 4])
    U = Zonotope.from_box(u_lo, u_up)
    print(U)
    print(U.alpha_to_control(alpha))



# Ref: https://apps.dtic.mil/sti/pdfs/ADA339188.pdf page 26

import math

import numpy as np

from cpsim import Simulator
from cpsim.controllers.PID import PID

# system dynamics
A = np.array([[-1.93 * math.pow(10, -2), 8.82, -32.2, -0.48],
              [-2.54 * math.pow(10, -4), -1.02, 0, 0.91],
              [0, 0, 0, 1],
              [2.95 * math.pow(10, -12), 0.82, 0, -1.08]])
B = np.array([[0.17], [-2.15 * math.pow(10, -3)], [0], [-0.18]])
C = np.array([0, 0, 57.3, 0]).reshape((4,))
D = np.array([0.0])

x_0 = np.array([[500.0], [0.0393], [0.0], [0.0393]]).reshape((4,))

# utils parameters
# KP = -1
# KI = 1.1
# KD = 0.00594
KP = -1.5
KI = -0.5
KD = 0.2
control_limit = {'lo': [-25], 'up': [25]}
#Ref: https://archive.siam.org/books/dc11/f16/Model.pdf page 6

class Controller:
    def __init__(self, dt):
        self.dt = dt
        self.pid = PID(KP, KI, KD, current_time=-dt)
        self.pid.setWindup(100)
        self.pid.setSampleTime(dt)
        self.set_control_limit(control_limit['lo'], control_limit['up'])

    def update(self, ref: np.ndarray, feedback_value: np.ndarray, current_time) -> np.ndarray:
        self.pid.set_reference(ref[0])
        cin = self.pid.update(feedback_value[0], current_time)
        return np.array([cin])

    def set_control_limit(self, control_lo, control_up):
        self.control_lo = control_lo
        self.control_up = control_up
        self.pid.set_control_limit(self.control_lo[0], self.control_up[0])

    def clear(self):
        self.pid.clear(current_time=-self.dt)


class F16(Simulator):
    """
            States: (4,)
                x[0]: velocity V[ft/sec]
                x[1]: angle of attack [rad]
                x[2]: pitch angle [rad]
                x[3]: pitch rate [rad/sec]
            Control Input: (1,)
                u[0]: elevator deflection [deg]
            Output:  (1,)
                y[0]: pitch angle * 57.3
                Output Feedback
            Controller: PID
            """

    def __init__(self, name, dt, max_index, noise=None):
        super().__init__('F16 ' + name, dt, max_index)
        self.linear(A, B, C, D)
        controller = Controller(dt)
        settings = {
            'init_state': x_0,
            'feedback_type': 'output',
            'controller': controller
        }
        if noise:
            settings['noise'] = noise
        self.sim_init(settings)


if __name__ == "__main__":
    max_index = 500
    dt = 0.02
    ref = [np.array([0.0872665 * 57.3])] * 501
    noise = {
        'process': {
            'type': 'white',
            'param': {'C': np.eye(4) * 0.0001}
        }
    }
    f16 = F16('test', dt, max_index, noise)
    for i in range(0, max_index + 1):
        assert f16.cur_index == i
        f16.update_current_ref(ref[i])
        # attack here
        f16.evolve()
    # print results
    import matplotlib.pyplot as plt

    t_arr = np.linspace(0, 10, max_index + 1)
    ref = [x[0] for x in f16.refs[:max_index + 1]]
    y_arr = [x[0] for x in f16.outputs[:max_index + 1]]

    plt.plot(t_arr, y_arr, t_arr, ref)
    plt.show()

    # u_arr = [x[0] for x in f16.inputs[:max_index + 1]]
    # plt.plot(t_arr, u_arr)
    # plt.show()



# Ref: Data-Driven Science and Engineering: Machine Learning, Dynamical Systems, and Control (Session 8.7, Page 300)

import numpy as np

from cpsim import Simulator
from cpsim.controllers.LQR import LQR

# parameters
m = 1  # mass of rob
M = 5  # mass of cart
L = 2  # length of rob
g = -10
d = 1  # dumping (friction)
b = 1  # pendulum up (b=1)


def inverted_pendulum(t, x, u, use_imath=False):
    u=u[0]
    if use_imath:
        from interval import imath
        Sx = imath.sin(x[2])
        Cx = imath.cos(x[2])
        D = m * L * L * (M + m * (1 - Cx * Cx))

        dx = [0,0,0,0]
        dx[0] = x[1]
        dx[1] = (1 / D) * (-m * m * L * L * g * Cx * Sx + m * L * L * (m * L * x[3] * x[3] * Sx - d * x[1])) + m * L * L * (
                1 / D) * u
        dx[2] = x[3]
        dx[3] = (1 / D) * ((m + M) * m * g * L * Sx - m * L * Cx * (m * L * x[3] * x[3] * Sx - d * x[1])) - m * L * Cx * (
                1 / D) * u
    else:
        Sx = np.sin(x[2])
        Cx = np.cos(x[2])
        D = m * L * L * (M + m * (1 - Cx * Cx))

        dx = np.zeros((4,))
        dx[0] = x[1]
        dx[1] = (1 / D) * (-m * m * L * L * g * Cx * Sx + m * L * L * (m * L * x[3] * x[3] * Sx - d * x[1])) + m * L * L * (
                1 / D) * u
        dx[2] = x[3]
        dx[3] = (1 / D) * ((m + M) * m * g * L * Sx - m * L * Cx * (m * L * x[3] * x[3] * Sx - d * x[1])) - m * L * Cx * (
                1 / D) * u
    return dx

def inverted_pendulum_imath(t, x, u):
    return inverted_pendulum(t=t,x=x,u=u, use_imath=True)

x_0 = np.array([-1, 0, np.pi + 0.1, 0])
control_limit = {
    'lo': np.array([-50]),
    'up': np.array([50])
}

# utils parameters
A = np.array([[0, 1, 0, 0],
              [0, -d / M, b * m * g / M, 0],
              [0, 0, 0, 1],
              [0, -b * d / (M * L), -b * (m + M) * g / (M * L), 0]])
B = np.array([[0], [1 / M], [0], [b * 1 / (M * L)]])

R = np.array([[0.0001]])
Q = np.eye(4)*10
Q[2,2] = 1000

class Controller:
    def __init__(self, dt, control_limit=None):
        self.lqr = LQR(A, B, Q, R)
        self.set_control_limit(control_limit['lo'], control_limit['up'])

    def update(self, ref: np.ndarray, feedback_value: np.ndarray, current_time) -> np.ndarray:
        self.lqr.set_reference(ref)
        cin = self.lqr.update(feedback_value, current_time)
        return cin
    def set_control_limit(self, control_lo, control_up):
        self.control_lo = control_lo
        self.control_up = control_up
        self.lqr.set_control_limit(self.control_lo[0], self.control_up[0])
    def clear(self):
        return


class InvertedPendulum(Simulator):
    """
    States: (4,)
        x[0]: location of cart
        x[1]: dx[0]
        x[2]: pendulum angle  (down:0, up:pi)
        x[3]: dx[1]
    Control Input: (1,)  [control_limit]
        u[0]: force on the cart
    Output: (4,)
        State Feedback
    Controller: LQR
    """

    def __init__(self, name, dt, max_index, noise=None):
        super().__init__('Inverted Pendulum ' + name, dt, max_index)
        self.nonlinear(ode=inverted_pendulum, n=4, m=1, p=4)
        controller = Controller(dt, control_limit)
        settings = {
            'init_state': x_0,
            'feedback_type': 'state',
            'controller': controller
        }
        if noise:
            settings['noise'] = noise
        self.sim_init(settings)


if __name__ == "__main__":
    max_index = 800
    dt = 0.02
    ref = [np.array([1, 0, np.pi, 0])] * (max_index+1)
    # bias attack example
    from cpsim import Attack
    bias = np.array([-1, 0, 0, 0])
    bias_attack = Attack('bias', bias, 300)
    ip = InvertedPendulum('test', dt, max_index)
    for i in range(0, max_index + 1):
        assert ip.cur_index == i
        ip.update_current_ref(ref[i])
        # attack here
        ip.cur_feedback = bias_attack.launch(ip.cur_feedback, ip.cur_index, ip.states)
        ip.evolve()
    # print results
    import matplotlib.pyplot as plt

    fig, ax = plt.subplots(3, 1)
    ax1, ax2, ax3 = ax
    t_arr = np.linspace(0, 10, max_index + 1)
    ref1 = [x[0] for x in ip.refs[:max_index + 1]]
    y1_arr = [x[0] for x in ip.outputs[:max_index + 1]]
    ax1.set_title('x0-location')
    ax1.plot(t_arr, y1_arr, t_arr, ref1)
    ref2 = [x[2] for x in ip.refs[:max_index + 1]]
    y2_arr = [x[2] for x in ip.outputs[:max_index + 1]]
    ax2.set_title('x2-angle')
    ax2.plot(t_arr, y2_arr, t_arr, ref2)
    u_arr = [x[0] for x in ip.inputs[:max_index + 1]]
    ax3.set_title('u-force')
    ax3.plot(t_arr, u_arr)
    print(u_arr)
    plt.show()



import matplotlib.pyplot as plt
import numpy as np


class HalfSpace:
    """
    normal vector - l
    The half space is defined as l^T x > b
    """
    def __init__(self, l: np.ndarray, b: [int, float]):
        self.l = l
        self.b = b
        self.dim = l.shape[0]

    def plot(self, x1, x2, fig=None):
        if self.dim != 2:
            return NotImplemented
        y1 = (self.b - x1 * self.l[0]) / self.l[1]
        y2 = (self.b - x2 * self.l[0]) / self.l[1]
        if fig is None:
            fig = plt.figure()
        X = [x1, x2]
        Y = [y1, y2]
        plt.plot(X, Y)
        if fig is None:
            plt.show()
        return fig




#Ref: https://ctms.engin.umich.edu/CTMS/index.php?example=AircraftPitch&section=ControlPID
import numpy as np

from cpsim import Simulator
from cpsim.controllers.PID import PID

# system dynamics
A = [[-0.313, 56.7, 0],
     [-0.0139, -0.426, 0],
     [0, 56.7, 0]]
B = [[0.232], [0.0203], [0]]
C = [[0, 0, 1]]
D = [[0]]

x_0 = np.array([0.0, 0.0, 0.0])

# utils parameters
KP = 1.13
KI = 0.0253
KD = 0.0
control_limit = {'lo': [-20], 'up': [20]}

class Controller:
    def __init__(self, dt):
        self.dt = dt
        self.pid = PID(KP, KI, KD, current_time=-dt)
        self.pid.setWindup(100)
        self.pid.setSampleTime(dt)
        self.set_control_limit(control_limit['lo'], control_limit['up'])

    def update(self, ref: np.ndarray, feedback_value: np.ndarray, current_time) -> np.ndarray:
        self.pid.set_reference(ref[0])
        cin = self.pid.update(feedback_value[0], current_time)
        return np.array([cin])

    def set_control_limit(self, control_lo, control_up):
        self.control_lo = control_lo
        self.control_up = control_up
        self.pid.set_control_limit(self.control_lo[0], self.control_up[0])

    def clear(self):
        self.pid.clear(current_time=-self.dt)


class AircraftPitch(Simulator):
    """
      States: (3,)
          x[0]: Angle of attack
          x[1]: Pitch rate
          x[2]: Pitch angle
      Control Input: (1,)
          u[0]: the elevator deflection angle
      Output:  (1,)
          y[0]: the pitch angle of the aircraft
          Output Feedback
      Controller: PID
      """
    def __init__(self, name, dt, max_index, noise=None):
        super().__init__('Aircraft Pitch ' + name, dt, max_index)
        self.linear(A, B, C)
        controller = Controller(dt)
        settings = {
            'init_state': x_0,
            'feedback_type': 'output',
            'controller': controller
        }
        if noise:
            settings['noise'] = noise
        self.sim_init(settings)


if __name__ == "__main__":
    max_index = 1500
    dt = 0.02
    ref = [np.array([0.2])] * 1501
    noise = {
        'process': {
            'type': 'white',
            'param': {'C': np.eye(3) * 0.00001}
        }
    }
    aircraft_pitch = AircraftPitch('test', dt, max_index, noise)
    for i in range(0, max_index + 1):
        assert aircraft_pitch.cur_index == i
        aircraft_pitch.update_current_ref(ref[i])
        # attack here
        aircraft_pitch.evolve()
    # print results
    import matplotlib.pyplot as plt

    t_arr = np.linspace(0, 10, max_index + 1)
    ref = [x[0] for x in aircraft_pitch.refs[:max_index + 1]]
    y_arr = [x[0] for x in aircraft_pitch.outputs[:max_index + 1]]

    plt.plot(t_arr, y_arr, t_arr, ref)
    plt.show()

    u_arr = [x[0] for x in aircraft_pitch.inputs[:max_index + 1]]
    plt.plot(t_arr, u_arr)



import math
from copy import deepcopy

import numpy as np


class Estimator:
    def __init__(self, Ad, Bd, max_k, epsilon=None, c=None):
        self.Ad = Ad
        self.Bd = Bd
        self.c = c
        self.n = self.Ad.shape[0]
        self.epsilon = epsilon
        self.max_k = max_k
        self.Ad_k = [np.eye(self.n)]
        for i in range(max_k):
            self.Ad_k.append(self.Ad_k[-1].dot(self.Ad))
        self.Ad_k_Bd = [i.dot(self.Bd) for i in self.Ad_k]
        if self.c is not None:
            self.Ad_k_c = [i.dot(self.c) for i in self.Ad_k]
        self.Ad_k_Ad_k_T = [i.dot(i.T) for i in self.Ad_k]
        self.epsilon_coef = []
        sqrt_term = np.zeros((self.n,))
        for k in range(max_k):
            for i in range(self.n):
                sqrt_term[i] += math.sqrt(self.Ad_k_Ad_k_T[k][i, i])
            self.epsilon_coef.append(deepcopy(sqrt_term))

    def estimate(self, x_a, us):
        # us: shape (length, m)
        k = len(us)
        assert k < self.max_k
        assert x_a.shape[0] == self.n
        assert self.epsilon is not None
        control_sum_term = np.zeros((self.n,))
        for j in range(k):
            if self.c is not None:
                control_sum_term += self.Ad_k_Bd[j] @ us[k - 1 - j] + self.Ad_k_c[j]
            else:
                control_sum_term += self.Ad_k_Bd[j] @ us[k - 1 - j]
        x_0 = self.Ad_k[k] @ x_a + control_sum_term
        e = np.ones((self.n,)) * self.epsilon * self.epsilon_coef[k]
        x_0_lo = x_0 - e
        x_0_up = x_0 + e
        return x_0_lo, x_0_up, x_0

    def estimate_wo_bound(self, x_a, us):
        # us: shape (length, m)
        k = len(us)
        assert k < self.max_k
        assert x_a.shape[0] == self.n
        control_sum_term = np.zeros((self.n,))
        for j in range(k):
            if self.c is not None:
                control_sum_term += self.Ad_k_Bd[j] @ us[k - 1 - j] + self.Ad_k_c[j]
            else:
                control_sum_term += self.Ad_k_Bd[j] @ us[k - 1 - j]
        x_0 = self.Ad_k[k] @ x_a + control_sum_term
        return x_0

    def get_deadline(self, x_a, safe_set_lo, safe_set_up, control: np.array, max_k):
        k = max_k
        breaked = False
        # print('utils', utils)
        for i in range(max_k):
            i += 1
            control_series = [control] * i
            # for j in range(i - 1):
            #     control_series = np.hstack((control_series, utils))
            # # print("control_ser=", control_series)
            x_0_lo, x_0_up, x_0 = self.estimate(x_a, control_series)
            for j in range(np.size(x_a, 0)):
                if safe_set_lo[j] < x_0_lo[j] < safe_set_up[j] and safe_set_lo[j] < x_0_up[j] < safe_set_up[j]:
                    pass
                else:
                    breaked = True
                    break
            if breaked:
                k = i - 1
                # print('x_0_up', x_0_up)
                break
        # print('x_0_up', x_0_up)
        return k


if __name__ == '__main__':
    from scipy.signal import StateSpace, lsim

    A = np.array([[-10, 1], [-0.02, -2]])
    B = np.array([[0], [2]])
    C = np.array([1, 0])
    sys = StateSpace(A, B, C, 0)
    sysd = sys.to_discrete(0.02)

    est = Estimator(sysd.A, sysd.B, 20, 1e-7)
    x_0 = np.array([0, 0])
    control_lst = np.array([[1]] * 10)
    x0_lo, x0_up, x0 = est.estimate(x_0, control_lst)
    print(x0_lo, x0_up)

    t = np.arange(0, 0.22, 0.02)
    t, y, x = lsim(sys, np.array([[1]] * 11), t)
    print(t, x)



#Ref: https://web.stanford.edu/class/archive/ee/ee392m/ee392m.1056/Lecture2_LinearSystems.pdf page 6
# Nenad Popovich. Lateral motion utils of boeing 747
# by using full-order observer. In 2019 5th International
# Conference on Control, Automation and Robotics (IC-
# CAR), pages 377–383, 2019.
import numpy as np

from cpsim import Simulator
from cpsim.controllers.PID import PID

# system dynamics
A = np.array([[0, 0, 1, 0, 0],
              [0, -0.0558, -0.9968, 0.0802, 0.0415],
              [0, 0.598, -0.115, -0.0318, 0],
              [0, -3.05, 0.388, -0.4650, 0],
              [0, 0, 0.0805, 1, 0]])
B = np.array([[0], [0.00729], [-0.475], [0.153], [0]])
C = np.array([[1, 0, 0, 0, 0]])

x_0 = np.array([10.0, 0.0, 0.0, 0.0, 0.0])

# utils parameters
R = np.array([[10]])
Q = np.eye(5)
# KP = 0
# KI = 0
# KD = 0
KP = -100
KI = 0
KD = 30
control_limit = {'lo': [-30], 'up': [30]}


class Controller:
    def __init__(self, dt):
        self.dt = dt
        self.pid = PID(KP, KI, KD, current_time=-dt)
        self.pid.setWindup(100)
        self.pid.setSampleTime(dt)
        self.set_control_limit(control_limit['lo'], control_limit['up'])

    def update(self, ref: np.ndarray, feedback_value: np.ndarray, current_time) -> np.ndarray:
        self.pid.set_reference(ref[0])
        cin = self.pid.update(feedback_value[0], current_time)
        return np.array([cin])

    def set_control_limit(self, control_lo, control_up):
        self.control_lo = control_lo
        self.control_up = control_up
        self.pid.set_control_limit(self.control_lo[0], self.control_up[0])

    def clear(self):
        self.pid.clear(current_time=-self.dt)

# class Controller:
#     def __init__(self, dt, control_limit=None):
#         self.lqr = LQR(A, B, Q, R)
#         # self.lqr.set_control_limit(control_limit['lo'], control_limit['up'])
#
#     def update(self, ref: np.ndarray, feedback_value: np.ndarray, current_time) -> np.ndarray:
#         self.lqr.set_reference(ref)
#         cin = self.lqr.update(feedback_value, current_time)
#         return cin

class Boeing(Simulator):
    """
              States: (5,)
                  x[0]: Yaw angle
                  x[1]: Side-slip angle
                  x[2]: Yaw rate
                  x[3]: roll rate
                  x[4]: roll angle
              Control Input: (1,)
                  u[0]: Rudder
              Output:  (2,)
                  y[0]: Yaw angle
                  Output Feedback
              Controller: PID
              """

    def __init__(self, name, dt, max_index, noise=None):
        super().__init__('Boeing' + name, dt, max_index)
        self.linear(A, B, C)
        controller = Controller(dt)
        settings = {
            'init_state': x_0,
            'feedback_type': 'output',
            'controller': controller
        }
        if noise:
            settings['noise'] = noise
        self.sim_init(settings)


if __name__ == "__main__":
    max_index = 2000
    dt = 0.02
    ref = [np.array([1])] * (max_index + 1)
    noise = {
        'process': {
            'type': 'white',
            'param': {'C': np.eye(5) * 0.01}
        }
    }
    boeing = Boeing('test', dt, max_index, noise)
    for i in range(0, max_index + 1):
        assert boeing.cur_index == i
        boeing.update_current_ref(ref[i])
        # attack here
        boeing.evolve()
    # print results
    import matplotlib.pyplot as plt

    t_arr = np.linspace(0, 10, max_index + 1)
    ref = [x[0] for x in boeing.refs[:max_index + 1]]
    y_arr = [x[0] for x in boeing.outputs[:max_index + 1]]

    plt.plot(t_arr, y_arr, t_arr, ref)

    plt.show()

    # u_arr = [x[0] for x in boeing.inputs[:max_index + 1]]
    # plt.plot(t_arr, u_arr)
    # plt.show()
